# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import itertools as it
import os
import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, Hashable, List, Optional, Tuple, cast

import networkx as nx  # type: ignore
import networkx.algorithms.isomorphism as ism  # type: ignore

import nncf
from nncf.common.utils.dot_file_rw import write_dot_graph
from nncf.parameters import ModelType
from nncf.parameters import TargetDevice


class Patterns:
    """
    Stores all layer patterns to be fused determined by hardware specific.
    This essence is used in the quantization algorithm.
    The operations in these patterns should be considered as a single
    during the quantization algorithm.
    """

    def __init__(self) -> None:
        self._patterns_dict: Dict[str, GraphPattern] = {}
        self._full_pattern_graph = GraphPattern()

    def register(self, pattern: "GraphPattern", name: str, match: bool = True) -> None:
        """
        Registers new pattern.

        :param pattern: pattern to be added
        :param name: name associated with the pattern
        :param match: whether should the pattern used as fussing pattern
        """
        if name in self._patterns_dict:
            raise KeyError("{} is already registered".format(name))
        self._patterns_dict[name] = pattern
        if match:
            self._full_pattern_graph.add_pattern_alternative(pattern)

    def get_full_pattern_graph(self) -> "GraphPattern":
        return self._full_pattern_graph

    def visualize_full_pattern_graph(self, path: str) -> None:
        self._full_pattern_graph.dump_graph(path)

    def visualize_all_patterns(self, dir_path: str) -> None:
        """
        Dump graphs of all registered patterns to dir_path
        """
        for patten_name, pattern in self._patterns_dict.items():
            pattern.dump_graph(os.path.join(dir_path, patten_name + ".dot"))

    def visualize_pattern(self, pattern_name: str, path: str) -> None:
        self._patterns_dict[pattern_name].dump_graph(os.path.join(path))


class GraphPattern:
    """
    Describes layer patterns in model's graph.
    This class is used in quantizer arrangement search algorithm, representing layer fusing patterns

    :param ANY_PATTERN_NODE_TYPE: Special node type, meaning any type inside the pattern.
    :param NON_PATTERN_NODE_TYPE: Special node type, meaning any type outside the pattern.
    """

    LABEL_ATTR = "label"
    METATYPE_ATTR = "type"
    NODE_TYPE_ATTR = "metatype"
    ANY_PATTERN_NODE_TYPE = "ANY_PATTERN_NODE"
    NON_PATTERN_NODE_TYPE = "NON_PATTERN_NODE"
    PATTERN_NODE_TO_EXCLUDE = "PATTERN_NODE_TO_EXCLUDE"

    def __init__(self) -> None:
        self._graph = nx.DiGraph()
        self._node_counter = 0

    def __add__(self, other: "GraphPattern") -> "GraphPattern":
        """
        Add DiGraph nodes of other to self and add edge between
        last node of self's graph and first node of other's graph.

        The first and the last nodes are found by nx.lexicographical_topological_sort().

        For more complex cases that are not covered by this function, use `join_patterns()`.

        :param other: GraphPattern that will be added.
        :return: resulted GraphPattern.
        """

        final_pattern = GraphPattern()
        for self_subgraph in self.get_weakly_connected_subgraphs():
            for other_subgraph in other.get_weakly_connected_subgraphs():
                # As this operation should output all graph combinations
                # It is essential to create copies of subgraphs and
                # add merge all possible connections
                # A: (a) (b)
                # B: (c) (d)
                #              (a_copy)  (a_copy)  (b_copy) (b_copy)
                # A + B ---->     |          |        |        |
                #              (c_copy)  (d_copy)  (c_copy) (d_copy)
                #
                subgraph_copy = final_pattern._unite_with_copy_of_graph(self_subgraph)
                other_subgraph_copy = final_pattern._unite_with_copy_of_graph(other_subgraph)
                final_pattern._add_edge_connected_subgraphs(subgraph_copy, other_subgraph_copy)

        return final_pattern

    def __or__(self, other: "GraphPattern") -> "GraphPattern":
        """
        Add other's DiGraph nodes to self's DiGraph as a new weakly connected components.
        It is a syntax sugar of 'add_pattern_alternative()'

        :param other: GraphPattern that will be added
        :return: resulted GraphPattern.
        """
        new_pattern = copy.deepcopy(self)
        new_pattern._unite_with_copy_of_graph(other.graph)
        return new_pattern

    def __eq__(self, other: object) -> bool:
        is_isomorphic: Callable[[Any, Any], bool] = ism.is_isomorphic
        return isinstance(other, GraphPattern) and is_isomorphic(self._graph, other.graph)

    @property
    def graph(self) -> nx.DiGraph:
        return self._graph

    def _unite_with_copy_of_graph(self, graph: nx.DiGraph) -> nx.DiGraph:
        """
        Creates a copy of 'graph', relabels node names according to self.node_counter
        and then unites relabeled graph with graph of 'self'.

        :param graph: graph, with which, self's graph will be united.
        :return: resulted graph.
        """
        mapping = {}
        for node in graph.nodes:
            new_node = self._node_counter
            mapping[node] = new_node
            self._node_counter += 1
        other_graph_copy = nx.relabel_nodes(graph, mapping, copy=True)
        self._graph = nx.union(self._graph, other_graph_copy)
        return other_graph_copy

    def _add_edge_connected_subgraphs(self, first_graph: nx.DiGraph, second_graph: nx.DiGraph) -> None:
        """
        Adds an edge between last node of 'first_graph' and first node of 'second_graph',
        which are found by nx.lexicographical_topological_sort().

        :param first_graph: the graph which will be traversed the first in the united graph.
        :param second_graph: the graph which will be traversed the second in the united graph.
        """
        self_graph = self._graph
        last_node_first_graph = list(nx.lexicographical_topological_sort(first_graph, key=int))[-1]
        assert first_graph.out_degree(last_node_first_graph) == 0
        first_node_second_graph = list(nx.lexicographical_topological_sort(second_graph, key=int))[0]
        assert second_graph.in_degree(first_node_second_graph) == 0

        # Special case when first node is ANY_PATTERN_NODE_TYPE or NON_PATTERN_NODE_TYPE
        if (
            GraphPattern.ANY_PATTERN_NODE_TYPE
            in second_graph.nodes[first_node_second_graph][GraphPattern.METATYPE_ATTR]
            or GraphPattern.NON_PATTERN_NODE_TYPE
            in second_graph.nodes[first_node_second_graph][GraphPattern.METATYPE_ATTR]
        ):
            successors = self_graph.successors(first_node_second_graph)
            new_edges = list(it.product([last_node_first_graph], successors))
            self_graph.add_edges_from(new_edges)
            self_graph.remove_node(first_node_second_graph)
        else:
            self_graph.add_edge(last_node_first_graph, first_node_second_graph)

    def add_pattern_alternative(self, other: "GraphPattern") -> None:
        """
        Adds 'other' pattern as a weakly connected component to 'self' pattern.

        :param other: GraphPattern that will be added
        """
        self._unite_with_copy_of_graph(other.graph)

    def join_patterns(self, other: "GraphPattern", edges: Optional[List[Tuple[Hashable, Hashable]]] = None) -> None:
        """
        Adds 'other' pattern to 'self' pattern and connect nodes from self to other specified by 'edges'.

        If edges is None, connect all weakly connected components of self and other by adding edges between
        the last nodes of every weakly component of self and the first nodes of every weakly component other.
        The first and the last nodes are found by nx.lexicographical_topological_sort().

        # A: (a) (b)
        # B: (c) (d)
        #              (a_copy)  (a_copy)  (b_copy) (b_copy)
        # A + B ---->     |          |        |        |
        #              (c_copy)  (d_copy)  (c_copy) (d_copy)
        #

        If other starts from a node with ANY_PATTERN_NODE_TYPE or NON_PATTERN_NODE_TYPE types,
        the input node of the other will be discarded from the final pattern.

        :param other: GraphPattern that will be added
        :param edges: List of edges between self and other graphs.
            Edges must begin at self and finish at other.
        """
        if edges is None:
            self._graph = (self + other).graph
        else:
            # Unite nodes
            other_graph_copy = copy.deepcopy(other.graph)
            node_mapping = {}
            for node_key in other_graph_copy.nodes:
                node_mapping[node_key] = self._node_counter
                self._node_counter += 1
            other_graph_copy = nx.relabel_nodes(other_graph_copy, node_mapping, copy=True)

            saved_graph = copy.deepcopy(self._graph)
            self._graph = nx.union(saved_graph, other_graph_copy)
            # Add edge/edges
            remapped_edges = []
            for edge in edges:
                new_edge = (edge[0], node_mapping[edge[1]])
                remapped_edges.append(new_edge)
            self._graph.add_edges_from(remapped_edges)

    def add_node(self, **attrs: Dict[str, Any]) -> int:
        if GraphPattern.METATYPE_ATTR in attrs and not isinstance(attrs[GraphPattern.METATYPE_ATTR], list):
            attrs[GraphPattern.METATYPE_ATTR] = cast(Any, [attrs[GraphPattern.METATYPE_ATTR]])
        self._graph.add_node(self._node_counter, **attrs)
        self._node_counter += 1
        return self._node_counter - 1

    def add_edge(self, u_name: str, v_name: str) -> None:
        self._graph.add_edge(u_name, v_name)

    def add_edges_from(self, ebunch_to_add: List[Any], **attr: Dict[str, Any]) -> None:
        self._graph.add_edges_from(ebunch_to_add, **attr)

    def get_weakly_connected_subgraphs(self) -> List[nx.DiGraph]:
        return [self._graph.subgraph(c) for c in nx.weakly_connected_components(self._graph)]

    def dump_graph(self, path: str) -> None:
        write_dot_graph(self._graph, pathlib.Path(path))


def merge_two_types_of_operations(first_op: Dict[str, Any], second_op: Dict[str, Any], label: str) -> Dict[str, Any]:
    if GraphPattern.METATYPE_ATTR in first_op and GraphPattern.METATYPE_ATTR in second_op:
        res = {GraphPattern.METATYPE_ATTR: first_op[GraphPattern.METATYPE_ATTR]}
        res[GraphPattern.METATYPE_ATTR].extend(second_op[GraphPattern.METATYPE_ATTR])
        res[GraphPattern.LABEL_ATTR] = label
        return res
    raise nncf.InternalError("Incorrect dicts of operations")


@dataclass
class PatternDesc:
    """
    Contains needed fields for the description of the pattern.

    :param name: Specific pattern name.
    :param devices: A field containing the list of devices
        for which this pattern should be taken into account when quantizing.
        None value means that this pattern is applicable to all devices.
    :param model_types: This field contains the list of the model types
        for which this pattern should be taken into account when quantizing.
        None value means that this pattern is applicable to all model types.
    """

    name: str
    devices: Optional[List[TargetDevice]] = None
    model_types: Optional[List[ModelType]] = None


class HWFusedPatternNames(Enum):
    """
    Describes the patterns that will be fused during integer execution
    and would not be quantized in compression pipeline.
    """

    # ATOMIC OPERATIONS
    L2_NORM = PatternDesc("l2_norm")
    MVN = PatternDesc("mvn")
    GELU = PatternDesc("gelu")

    # BLOCK PATTERNS
    ADD_SCALE_SHIFT_OUTPUT = PatternDesc("add_scale_shift_output")
    BATCH_INDEX = PatternDesc("batch_index")
    LINEAR_WITH_BIAS = PatternDesc("linear_with_bias")
    MVN_SCALE_SHIFT = PatternDesc("mvn_scale_shift")
    NORMALIZE_L2_MULTIPLY = PatternDesc("normalize_l2_multiply")
    SCALE_SHIFT = PatternDesc("scale_shift")
    SHIFT_SCALE = PatternDesc("shift_scale")
    SOFTMAX_DIV = PatternDesc("softmax_div")

    # ACTIVATIONS
    HSWISH_ACTIVATION = PatternDesc("hswish_activation")
    HSWISH_ACTIVATION_V2 = PatternDesc("hswish_activation_v2")
    HSWISH_ACTIVATION_WITHOUT_DENOMINATOR = PatternDesc("hswish_activation_without_denominator")
    SOFTMAX = PatternDesc("softmax")
    SWISH_WITH_HARD_SIGMOID = PatternDesc("swish_with_hard_sigmoid")
    SWISH_WITH_SIGMOID = PatternDesc("swish_with_sigmoid")

    # INPUT PROCESSING
    INPUT_CONVERT_TRANSPOSE_PROCESSING = PatternDesc("input_convert_transpose_processing")
    INPUT_CONVERT_TRANSPOSE_REVERSE_ADD = PatternDesc("input_convert_transpose_reverse_add")
    INPUT_CONVERT_TRANSPOSE_REVERSE_SCALE_SHIFT = PatternDesc("input_convert_transpose_reverse_scale_shift")
    INPUT_CONVERT_TRANSPOSE_SCALE_SHIFT = PatternDesc("input_convert_transpose_scale_shift")
    INPUT_PROCESSING = PatternDesc("input_processing")
    INPUT_REVERSE_ADD = PatternDesc("input_reverse_add")
    INPUT_REVERSE_SCALE_SHIFT = PatternDesc("input_reverse_scale_shift")
    INPUT_SCALE_SHIFT = PatternDesc("input_scale_shift")
    INPUT_SHIFT_SCALE = PatternDesc("input_shift_scale")
    INPUT_TRANSPOSE_PROCESSING = PatternDesc("input_transpose_processing")
    INPUT_TRANSPOSE_REVERSE_ADD = PatternDesc("input_transpose_reverse_add")
    INPUT_TRANSPOSE_SCALE_SHIFT = PatternDesc("input_transpose_scale_shift")

    # COMBINATIONS
    ACTIVATIONS_BATCH_NORM = PatternDesc("activations_batch_norm")
    ACTIVATIONS_SCALE_SHIFT = PatternDesc("activations_scale_shift")
    ARITHMETIC_ACTIVATIONS = PatternDesc("arithmetic_activations")
    ARITHMETIC_ACTIVATIONS_BATCH_NORM = PatternDesc("arithmetic_activations_batch_norm")
    # StyleGan2
    ARITHMETIC_ACTIVATIONS_ARITHMETIC = PatternDesc("arithmetic_activations_arithmetic")
    ARITHMETIC_ACTIVATIONS_SCALE_SHIFT = PatternDesc("arithmetic_activations_scale_shift")
    ARITHMETIC_BATCH_NORM = PatternDesc("arithmetic_batch_norm")
    ARITHMETIC_BATCH_NORM_ACTIVATIONS = PatternDesc("arithmetic_batch_norm_activations")
    ARITHMETIC_SCALE_SHIFT = PatternDesc("arithmetic_scale_shift")
    ARITHMETIC_SCALE_SHIFT_ACTIVATIONS = PatternDesc("arithmetic_scale_shift_activations")
    BATCH_NORM_ACTIVATIONS = PatternDesc("batch_norm_activations")
    BATCH_NORM_SCALE_SHIFT_ACTIVATIONS = PatternDesc("batch_norm_scale_shift_activations")
    GROUP_NORM_RELU = PatternDesc("group_norm_relu")
    LINEAR_ACTIVATIONS = PatternDesc("linear_activations")
    LINEAR_ACTIVATIONS_BATCH_NORM = PatternDesc("linear_activations_batch_norm")
    LINEAR_ACTIVATIONS_SCALE_SHIFT = PatternDesc("linear_activations_scale_shift")
    LINEAR_ARITHMETIC = PatternDesc("linear_arithmetic")
    LINEAR_SHIFT_SCALE = PatternDesc("linear_shift_scale")
    LINEAR_ARITHMETIC_ACTIVATIONS = PatternDesc("linear_arithmetic_activations")
    # Found in PicoDet models
    LINEAR_ARITHMETIC_ACTIVATIONS_ARITHMETIC = PatternDesc("linear_arithmetic_activations_arithmetic")
    LINEAR_BATCH_NORM = PatternDesc("linear_batch_norm")
    LINEAR_BATCH_NORM_ACTIVATIONS = PatternDesc("linear_batch_norm_activations")
    # MaskRCNN_Resnet_Atrous
    LINEAR_BATCH_TO_SPACE_SCALE_SHIFT_ACTIVATIONS = PatternDesc("linear_batch_to_space_scale_shift_activations")
    LINEAR_BATCH_TO_SPACE_ARITHMETIC_ACTIVATIONS = PatternDesc("linear_batch_to_space_arithmetic_activations")
    LINEAR_BATCH_NORM_SCALE_SHIFT_ACTIVATIONS = PatternDesc("linear_batch_norm_scale_shift_activations")
    LINEAR_SCALE_SHIFT_ACTIVATIONS = PatternDesc("linear_scale_shift_activations")
    LINEAR_CONST_MULTIPLY = PatternDesc("linear_const_multiply")
    LINEAR_SQUEEZE_ACTIVATIONS = PatternDesc("linear_squeeze_activations")
    LINEAR_SQUEEZE_ARITHMETIC_ACTIVATIONS = PatternDesc("linear_squeeze_arithmetic_activations")
    LINEAR_ACTIVATIONS_UNSQUEEZE_BN_SQUEEZE = PatternDesc("linear_activations_unsqueeze_bn_squeeze")
    SCALE_SHIFT_ACTIVATIONS = PatternDesc("scale_shift_activations")
    MVN_SCALE_SHIFT_ACTIVATIONS = PatternDesc("mvn_scale_shift_activations")

    # DEVICE PATTERNS
    HSWISH_ACTIVATION_CLAMP_MULTIPLY = PatternDesc(
        "hswish_activation_clamp_multiply",
        devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU, TargetDevice.NPU],
    )
    LINEAR_SCALE_SHIFT = PatternDesc(
        "linear_scale_shift", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_BIASED_SCALE_SHIFT = PatternDesc(
        "linear_biased_scale_shift", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_ACTIVATION_SCALE_SHIFT = PatternDesc(
        "linear_activation_scale_shift", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_BIASED_ACTIVATION_SCALE_SHIFT = PatternDesc(
        "linear_biased_activation_scale_shift", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_ELEMENTWISE = PatternDesc(
        "linear_elementwise", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_BIASED_ELEMENTWISE = PatternDesc(
        "linear_biased_elementwise", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_ACTIVATION_ELEMENTWISE = PatternDesc(
        "linear_activation_elementwise", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )
    LINEAR_BIASED_ACTIVATION_ELEMENTWISE = PatternDesc(
        "linear_biased_activation_elementwise", devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU]
    )


class IgnoredPatternNames(Enum):
    """
    Describes the patterns, which nodes should be ignored during FakeQuantize placement.
    """

    MULTIHEAD_ATTENTION_OUTPUT = PatternDesc(
        "multihead_attention_output",
        model_types=[ModelType.TRANSFORMER],
        devices=[TargetDevice.ANY, TargetDevice.CPU, TargetDevice.GPU, TargetDevice.NPU],
    )
    SE_BLOCK = PatternDesc("se_block")
    FC_BN_HSWISH_ACTIVATION = PatternDesc("fc_bn_hswish_activation")
    EQUAL_LOGICALNOT = PatternDesc("equal_logicalnot")
    ROPE = PatternDesc("rope", model_types=[ModelType.TRANSFORMER])



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List, Optional, Tuple

import torch

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.logging import nncf_logger
from nncf.torch.graph.graph import NNCFNode
from nncf.torch.layers import NNCF_GENERAL_CONV_MODULES_DICT
from nncf.torch.layers import NNCF_LINEAR_MODULES_DICT
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.tensor import PTNNCFTensor


def get_bn_node_for_conv(graph: NNCFGraph, conv_node: NNCFNode) -> Optional[NNCFNode]:
    successors = graph.get_next_nodes(conv_node)
    for succ in successors:
        if succ.node_type == "batch_norm":
            return succ
    return None


def get_bn_for_conv_node_by_name(target_model: NNCFNetwork, conv_node_name: NNCFNodeName) -> Optional[torch.nn.Module]:
    """
    Returns a batch norm module in target_model that corresponds immediately following a given
    convolution node in the model's NNCFGraph representation.
    :param target_model: NNCFNetwork to work with
    :param module_scope:
    :return: batch norm module
    """
    graph = target_model.nncf.get_original_graph()
    conv_node = graph.get_node_by_name(conv_node_name)
    bn_node = get_bn_node_for_conv(graph, conv_node)
    if bn_node is None:
        return None
    bn_module = target_model.nncf.get_containing_module(bn_node.node_name)
    return bn_module


def init_output_masks_in_graph(graph: NNCFGraph, nodes: List):
    """
    Initialize masks in graph for mask propagation algorithm

    :param graph: NNCFNetwork
    :param nodes: list with pruned nodes
    """
    for node in graph.get_all_nodes():
        node.attributes.pop("output_mask", None)

    for minfo in nodes:
        mask = minfo.operand.binary_filter_pruning_mask
        nncf_node = graph.get_node_by_id(minfo.nncf_node_id)
        nncf_node.attributes["output_mask"] = PTNNCFTensor(mask)


def _calculate_output_shape(graph: NNCFGraph, node: NNCFNode) -> Tuple[int, ...]:
    """
    Calculates output shape of convolution layer by input edge.

    :param graph: the model graph
    :param node: node from NNCF graph
    :return: output shape
    """
    in_edge = graph.get_input_edges(node)[0]
    shape = list(in_edge.tensor_shape)
    attrs = node.layer_attributes

    if isinstance(attrs, ConvolutionLayerAttributes):
        shape = shape[2:]
        for i, _ in enumerate(shape):
            if attrs.transpose:
                shape[i] = (shape[i] - 1) * attrs.stride[i] - 2 * attrs.padding_values[i] + attrs.kernel_size[i]
            else:
                shape[i] = (shape[i] + 2 * attrs.padding_values[i] - attrs.kernel_size[i]) // attrs.stride[i] + 1
    elif isinstance(attrs, LinearLayerAttributes):
        shape = shape[:-1] + [attrs.out_features]
    else:
        raise nncf.ValidationError(f"Unexpected node type {node.node_type} is fed to _calculate_output_shape")
    return tuple(shape)


def collect_output_shapes(graph: NNCFGraph) -> Dict[NNCFNodeName, List[int]]:
    """
    Collects output dimension shapes for convolutions and fully connected layers
    from the connected edges in the NNCFGraph.

    :param graph: NNCFGraph.
    :return: Dictionary of output dimension shapes. E.g {node_name: (height, width)}
    """
    modules_out_shapes = {}
    output_shape_collecting_info = [
        (NNCF_GENERAL_CONV_MODULES_DICT, slice(2, None)),
        (NNCF_LINEAR_MODULES_DICT, slice(None)),
    ]
    for nncf_module_type, shape_slice in output_shape_collecting_info:
        for node in graph.get_nodes_by_types([v.op_func_name for v in nncf_module_type]):
            output_edges = graph.get_output_edges(node)
            if output_edges:
                out_edge = output_edges[0]
                out_shape = out_edge.tensor_shape[shape_slice]
            else:
                # For disconnected NNCFGraph when node have no output edge
                out_shape = _calculate_output_shape(graph, node)
                nncf_logger.debug(f"Node {node.node_name} has no output edge in NNCFGraph")
            modules_out_shapes[node.node_name] = out_shape

    return modules_out_shapes



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Tuple, TypeVar

import numpy as np

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase

TModel = TypeVar("TModel")


@dataclass
class LayoutDescriptor:
    """
    Container to store convolutional and linear layers layout information.
    """

    conv_weight_out_channels_dim: int
    conv_weight_in_channels_dim: int
    bias_channels_dim: int


class ChannelAlignmentAlgoBackend:
    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> TargetPoint:
        """
        Returns backend-specific target point.

        :param target_type: Type of the location that should be modified.
        :param target_node_name: Name of the located node.
        :param port_id: id of the port for the statistics distribution.
        :return: Backend-specific TargetPoint.
        """

    @staticmethod
    @abstractmethod
    def get_bias_value(node: NNCFNode, model: TModel, nncf_graph: NNCFGraph) -> np.ndarray:
        """
        Returns bias value in the NumPy format of provided node.

        :param node: Node of NNCFGraph with bias value.
        :param model: Backend-specific model for the initializer finding.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Bias value in the NumPy format.
        """

    @staticmethod
    @abstractmethod
    def get_weight_value(node: NNCFNode, model: TModel, port_id: int) -> np.ndarray:
        """
        Returns bias value in the NumPy format of provided node.

        :param node: Node of NNCFGraph with bias value.
        :param model: Backend-specific model for the initializer finding.
        :param nncf_graph: NNCFGraph instance with the node.
        :return: Bias value in the NumPy format.
        """

    @staticmethod
    @abstractmethod
    def get_activation_port_ids_for_node(node: NNCFNode) -> Tuple[int, int]:
        """
        Returns Input Port ID and Output Port ID corresponding to activation input and output edges for
        the node.
        Supports only nodes that could have bias value.

        :param node: Node of NNCFGraph with bias value.
        """

    @staticmethod
    @abstractmethod
    def get_weights_port_ids_for_node(node: NNCFNode) -> Tuple[int, int]:
        """
        Returns Input Port ID and Output Port ID corresponding to node weights input port id and
        constant output port id the node.

        :param node: Node of NNCFGraph.
        """

    @staticmethod
    @abstractmethod
    def get_statistic_collector(
        reduction_axes, q: float, num_samples: int, inplace: bool
    ) -> TensorStatisticCollectorBase:
        """
        Get backend-specific tensor collector that collects medians of minimal and maximal quantiles.

        :param reduction_axes: Target reduction axes for the reduction.
        :param q: Minimal quantile for the tensor collector.
        :param num_samples: Num samples to collect by the tensor collector.
        :param inplace: Should statistic be calculated inplace or out of place.
        :return: Backend-specific tensor collector that collects medians of minimal and maximal quantiles.
        """

    @staticmethod
    @abstractmethod
    def get_conv_layer_attributes(node: NNCFNode) -> ConvolutionLayerAttributes:
        """
        Returns convolutional layer attributes of given node if they are present and None otherwise.
        :param node: NNCFNode to take convolutional layer attributes from.
        :return: Convolutional layer attributes of given node if they are present and None otherwise
        """

    @staticmethod
    @abstractmethod
    def get_dims_descriptor(node: NNCFNode) -> LayoutDescriptor:
        """
        Return weights layout descriptor of the given node if it is possible and None otherwise.
        Only convolutional and linear nodes are supported.

        :param node: NNCFNode to get layout descriptor from.
        :return: Weights layout descriptor of the given node if it is possible and None otherwise.
        """

    @staticmethod
    @abstractmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Checks if the node has a bias or not.

        :param node: The node to check.
        :param nncf_graph: The NNCF graph.
        :return: True` if `node` corresponds to the operation with bias
            (bias is added to the output tensor of that operation), `False` otherwise.
        """

    @staticmethod
    @abstractmethod
    def create_bias_tensor(node: NNCFNode, nncf_graph: NNCFGraph, value: Any) -> np.ndarray:
        """
        Creates bias value constant array filled by given value.

        :param node: NNCFNode to add bias to.
        :param nncf_graph: Target NNCFgraph.
        :param value: Value to fill bias constant array.
        :return: Bias value constant array filled by given value.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf import NNCFConfig


def is_accuracy_aware_training(nncf_config: NNCFConfig) -> bool:
    if nncf_config.get("accuracy_aware_training") is not None:
        return True
    return False


def is_experimental_quantization(nncf_config: NNCFConfig) -> bool:
    compression_section = nncf_config.get("compression", [])
    algorithms_names = []

    if isinstance(compression_section, dict):
        algorithms_names.append(compression_section["algorithm"])

    if isinstance(compression_section, list):
        for sec in compression_section:
            algorithms_names.append(sec["algorithm"])

    return "experimental_quantization" in algorithms_names



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns.patterns import GraphPattern
from nncf.common.graph.patterns.patterns import IgnoredPatternNames
from nncf.common.utils.registry import Registry
from nncf.onnx.graph.metatypes import onnx_metatypes as om
from nncf.onnx.graph.metatypes.groups import MATMUL_METATYPES
from nncf.onnx.hardware.fused_patterns import atomic_activations_operations

ONNX_IGNORED_PATTERNS = Registry("IGNORED_PATTERNS")


def _add_softmax_matmul(pattern: GraphPattern) -> None:
    #       SOFTMAX  RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #           \              /
    #            \            /
    #             \          /
    #              \        /
    #               \      /
    #                MATMUL
    branch_matmul_nodes = [
        om.ONNXReshapeMetatype,
        om.ONNXTransposeMetatype,
        om.ONNXGatherMetatype,
        om.ONNXSqueezeMetatype,
        om.ONNXConcatMetatype,
    ]
    softmax = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: om.ONNXSoftmaxMetatype}
    )
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: MATMUL_METATYPES})
    matmul_branch_nodes = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT",
            GraphPattern.METATYPE_ATTR: branch_matmul_nodes,
        }
    )
    pattern.add_edge(softmax, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)


def _add_softmax_reshape_matmul(pattern: GraphPattern) -> None:
    #       SOFTMAX
    #           \
    #            \
    #             \
    #             RESHAPE   RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #                 \                 /
    #                  \               /
    #                   \             /
    #                    \           /
    #                     \         /
    #                      \       /
    #                        MATMUL
    branch_matmul_nodes = [
        om.ONNXReshapeMetatype,
        om.ONNXTransposeMetatype,
        om.ONNXGatherMetatype,
        om.ONNXSqueezeMetatype,
        om.ONNXConcatMetatype,
    ]
    softmax = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: om.ONNXSoftmaxMetatype}
    )
    reshape = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "RESHAPE", GraphPattern.METATYPE_ATTR: om.ONNXReshapeMetatype}
    )
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: MATMUL_METATYPES})
    matmul_branch_nodes = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT",
            GraphPattern.METATYPE_ATTR: branch_matmul_nodes,
        }
    )
    pattern.add_edge(softmax, reshape)
    pattern.add_edge(reshape, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)


@ONNX_IGNORED_PATTERNS.register(IgnoredPatternNames.MULTIHEAD_ATTENTION_OUTPUT)
def create_multihead_attention_output() -> GraphPattern:
    pattern = GraphPattern()
    _add_softmax_matmul(pattern)
    _add_softmax_reshape_matmul(pattern)
    return pattern


@ONNX_IGNORED_PATTERNS.register(IgnoredPatternNames.SE_BLOCK)
def create_se_block() -> GraphPattern:
    #       NON_PATTERN_NODE--------------
    #    (PATTERN_NODE_TO_EXCLUDE)       |
    #              |                     |
    #   REDUCE_MEAN||GLOBAL_AVERAGE_POOL |
    #              |                     |
    #         CONVOLUTION                |
    #              |                     |
    #          ACTIVATION                |
    #              |                     |
    #         CONVOLUTION                |
    #              |                     |
    #       SIGMOID||HARDSIGMOID         |
    #              |                     |
    #             MUL---------------------
    #   (PATTERN_NODE_TO_EXCLUDE)
    pattern = GraphPattern()
    non_pattern_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    reduce_node_or_global_average_pool = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "REDUCE_MEAN || GLOBAL_AVERAGE_POOL",
            GraphPattern.METATYPE_ATTR: [om.ONNXReduceMeanMetatype, om.ONNXGlobalAveragePoolMetatype],
            GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
        }
    )
    conv_node_1 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONVOLUTION", GraphPattern.METATYPE_ATTR: om.ONNXConvolutionMetatype}
    )

    pattern.add_edge(non_pattern_node, reduce_node_or_global_average_pool)
    pattern.add_edge(reduce_node_or_global_average_pool, conv_node_1)
    pattern.join_patterns(atomic_activations_operations())

    rest_pattern = GraphPattern()
    conv_node_2 = rest_pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONVOLUTION", GraphPattern.METATYPE_ATTR: om.ONNXConvolutionMetatype}
    )

    sigmoid_node = rest_pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "SIGMOID",
            GraphPattern.METATYPE_ATTR: [om.ONNXSigmoidMetatype, om.ONNXHardSigmoidMetatype],
        }
    )
    multiply_node = rest_pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "LAST_MULTIPLY",
            GraphPattern.METATYPE_ATTR: om.ONNXMulLayerMetatype,
            GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
        }
    )
    rest_pattern.add_edge(conv_node_2, sigmoid_node)
    rest_pattern.add_edge(sigmoid_node, multiply_node)
    pattern.join_patterns(rest_pattern)
    # Connect all NON_PATTERN_NODE with all MULTIPLY
    for component in pattern.get_weakly_connected_subgraphs():
        for node_id, attrs in component.nodes(data=True):
            if attrs[GraphPattern.LABEL_ATTR] == "NON_PATTERN_NODE":
                non_pattern_node = node_id
            if attrs[GraphPattern.LABEL_ATTR] == "LAST_MULTIPLY":
                multiply_node = node_id
        pattern.add_edge(non_pattern_node, multiply_node)
    return pattern


@ONNX_IGNORED_PATTERNS.register(IgnoredPatternNames.ROPE)
def create_rope() -> GraphPattern:
    pattern = GraphPattern()
    matmul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: om.ONNXMatMulMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.ONNXTransposeMetatype}
    )
    concat_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.ONNXConcatMetatype}
    )
    cos_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "COS", GraphPattern.METATYPE_ATTR: om.ONNXCosMetatype})
    sin_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SIN", GraphPattern.METATYPE_ATTR: om.ONNXSinMetatype})

    pattern.add_edge(matmul_node, transpose_node)
    pattern.add_edge(transpose_node, concat_node)
    pattern.add_edge(concat_node, cos_node)
    pattern.add_edge(concat_node, sin_node)
    return pattern



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import numpy as np
import openvino.runtime as ov
from openvino import Type
from openvino.properties.hint import inference_precision

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.openvino.engine import OVCompiledModelEngine
from nncf.openvino.graph.layer_attributes import OVLayerAttributes
from nncf.openvino.graph.metatypes.groups import CONSTANT_OPERATIONS
from nncf.openvino.graph.metatypes.groups import FAKE_QUANTIZE_OPERATIONS
from nncf.openvino.graph.metatypes.groups import INPUTS_QUANTIZABLE_OPERATIONS
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_WEIGHTS
from nncf.openvino.graph.metatypes.groups import QUANTIZE_AGNOSTIC_OPERATIONS
from nncf.openvino.graph.metatypes.groups import SHAPEOF_OPERATIONS
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConcatMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVOpMetatype
from nncf.openvino.graph.model_utils import get_start_nodes_for_activation_path_tracing
from nncf.openvino.graph.model_utils import model_has_state
from nncf.openvino.graph.node_utils import get_bias_value
from nncf.openvino.graph.node_utils import get_weight_value
from nncf.openvino.graph.node_utils import is_node_with_bias
from nncf.quantization.algorithms.accuracy_control.backend import AccuracyControlAlgoBackend
from nncf.quantization.algorithms.accuracy_control.backend import PreparedModel


class OVPreparedModel(PreparedModel):
    """
    Implementation of the `PreparedModel` for OpenVINO backend.
    """

    def __init__(self, model: ov.Model, use_fp32_precision: bool = True):
        self._stateful = model_has_state(model)
        config = None
        if use_fp32_precision:
            config = {inference_precision: Type.f32}
        self._compiled_model = ov.compile_model(model, device_name="CPU", config=config)
        self._engine = None

    @property
    def model_for_inference(self) -> ov.CompiledModel:
        return self._compiled_model

    @property
    def engine(self) -> OVCompiledModelEngine:
        if self._engine is None:
            self._engine = OVCompiledModelEngine(self._compiled_model, self._stateful)
        return self._engine


class OVAccuracyControlAlgoBackend(AccuracyControlAlgoBackend):
    """
    Implementation of the `AccuracyControlAlgoBackend` for OpenVINO backend.
    """

    # Metatypes

    @staticmethod
    def get_op_with_weights_metatypes() -> List[OVOpMetatype]:
        return OPERATIONS_WITH_WEIGHTS

    @staticmethod
    def get_quantizer_metatypes() -> List[OVOpMetatype]:
        return FAKE_QUANTIZE_OPERATIONS

    @staticmethod
    def get_const_metatypes() -> List[OVOpMetatype]:
        return CONSTANT_OPERATIONS

    @staticmethod
    def get_quantizable_metatypes() -> List[OVOpMetatype]:
        return INPUTS_QUANTIZABLE_OPERATIONS

    @staticmethod
    def get_quantize_agnostic_metatypes() -> List[OVOpMetatype]:
        return QUANTIZE_AGNOSTIC_OPERATIONS + [OVConcatMetatype]

    @staticmethod
    def get_shapeof_metatypes() -> List[OVOpMetatype]:
        return SHAPEOF_OPERATIONS

    @staticmethod
    def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return get_start_nodes_for_activation_path_tracing(nncf_graph)

    # Manipulations with bias value and weights

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node, nncf_graph)

    @staticmethod
    def is_node_with_weight(node: NNCFNode) -> bool:
        return node.metatype in OPERATIONS_WITH_WEIGHTS and isinstance(node.layer_attributes, OVLayerAttributes)

    @staticmethod
    def get_bias_value(node_with_bias: NNCFNode, nncf_graph: NNCFGraph, model: ov.Model) -> np.ndarray:
        return get_bias_value(node_with_bias, nncf_graph, model)

    @staticmethod
    def get_weight_value(node_with_weight: NNCFNode, model: ov.Model, port_id: int) -> np.ndarray:
        return get_weight_value(node_with_weight, model, port_id)

    @staticmethod
    def get_weight_tensor_port_ids(node: NNCFNode) -> List[Optional[int]]:
        return node.layer_attributes.get_const_port_ids()

    @staticmethod
    def get_model_size(model: ov.Model) -> int:
        model_size = 0
        for op in model.get_ops():
            if op.get_type_name() == "Constant":
                model_size += op.data.nbytes

        return model_size



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, TypeVar, Union

TObj = TypeVar("TObj", bound=Union[Callable[..., Any], type])

API_MARKER_ATTR = "_nncf_api_marker"
CANONICAL_ALIAS_ATTR = "_nncf_canonical_alias"


def api(canonical_alias: str = None) -> Callable[[TObj], TObj]:
    """
    Decorator function used to mark a object as an API.

    Example:
        @api(canonical_alias="alias")
        class Class:
            pass

        @api(canonical_alias="alias")
        def function():
            pass

    :param canonical_alias: The canonical alias for the API class.
    """

    def decorator(obj: TObj) -> TObj:
        setattr(obj, API_MARKER_ATTR, obj.__name__)
        if canonical_alias is not None:
            setattr(obj, CANONICAL_ALIAS_ATTR, canonical_alias)
        return obj

    return decorator



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Dict, List, Optional, Set, Tuple, TypeVar

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.hardware.config import HWConfig
from nncf.common.quantization.structs import QuantizerConfig
from nncf.experimental.common.tensor_statistics.collectors import TensorReducerBase
from nncf.parameters import ModelType
from nncf.parameters import TargetDevice
from nncf.quantization.fake_quantize import FakeConvertParameters
from nncf.quantization.fake_quantize import FakeQuantizeParameters
from nncf.quantization.range_estimator import StatisticsType

TModel = TypeVar("TModel")


class MinMaxAlgoBackend(ABC):
    @property
    @abstractmethod
    def preserved_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for backend-specific metatypes that require preserving float subgraphs
        when removing the ShapeOf subgraph.
        """

    @property
    @abstractmethod
    def mat_mul_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific MatMul metatypes.
        """

    @property
    @abstractmethod
    def post_processing_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific post-processing metatypes (NonMaximumSupression, TopK, etc.).
        """

    @property
    @abstractmethod
    def conv_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific Convolution metatypes.
        """

    @property
    @abstractmethod
    def shapeof_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific ShapeOf metatypes.
        """

    @property
    @abstractmethod
    def dropout_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific Dropout metatypes.
        """

    @property
    @abstractmethod
    def elementwise_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific Elementwises metatypes.
        """

    @property
    @abstractmethod
    def overflow_fix_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific metatypes for which overflow_fix is applicable.
        """

    @property
    @abstractmethod
    def add_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific metatypes that also can be interpreted as Add layer.
        """

    @property
    @abstractmethod
    def group_conv_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific Grouped Convolution metatypes.
        """

    @property
    @abstractmethod
    def scaled_dot_product_attention_metatypes(self) -> List[OperatorMetatype]:
        """
        Property for the backend-specific Scaled Dot Product Attention metatypes.
        """

    @property
    @abstractmethod
    def scales_unification_map(self) -> Dict[OperatorMetatype, OperatorMetatype]:
        """
        Property for the backend-specific metatypes that produces quantizers that might be unified.
        """

    @property
    @abstractmethod
    def hw_config(self) -> HWConfig:
        """
        Property for the hardware backend-specific configuration.
        """

    @property
    @abstractmethod
    def quant_trait_op_dict(self) -> Dict[int, OperatorMetatype]:
        """
        Property for the backend-specific dictionary that contains QuantizationTrait-specific metatypes.
        """

    @property
    @abstractmethod
    def reducer_map(self) -> Dict[StatisticsType, TensorReducerBase]:
        """
        Property for the backend-specific dictionary that contains backend-specific tensor reducers.
        """

    @property
    @abstractmethod
    def supports_inplace_statistics(self) -> bool:
        """
        Property for the backend-specific flag that specifies whether the backend supports inplace statistics.
        """

    @staticmethod
    @abstractmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> TargetPoint:
        """
        Returns backend-specific target point.

        :param target_type: Type of the location that should be modified.
        :param target_node_name: Name of the located node.
        :param port_id: Port ID of the tensor for the statistics distribution.
        :return: Backend-specific TargetPoint.
        """

    @staticmethod
    @abstractmethod
    def create_quantizer_insertion_command(
        nncf_graph: NNCFGraph,
        target_point: TargetPoint,
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> TransformationCommand:
        """
        Returns backend-specific quantizer insertion command.

        :param nncf_graph: NNCFGraph to get input/output shapes for the target point.
        :param target_point: Target location for the quantizer insertion.
        :param quantizer_config: QuantizerConfig instance for the current layer.
        :param parameters: FakeQuantizeParameters to calculate activation quantization parameters.
        :return: Backend-specific TransformationCommand for the quantizer insertion operation.
        """

    @staticmethod
    @abstractmethod
    def create_unified_scales_quantizers_insertion_commands(
        nncf_graph: NNCFGraph,
        target_points: List[TargetPoint],
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> List[TransformationCommand]:
        """
        Returns backend-specific unified scales quantizers insertion commands.

        :param nncf_graph: NNCFGraph to get input/output shapes for the target point.
        :param target_points: List of target locations for the quantizers insertion.
        :param quantizer_config: QuantizerConfig instance for the current layer.
        :param parameters: FakeQuantizeParameters to calculate activation quantization parameters.
        :return: List of backend-specific TransformationCommands
            for the quantizers with unified scales insertion operations.
        """

    @staticmethod
    @abstractmethod
    def create_convert_insertion_command(
        target_point: TargetPoint,
        parameters: FakeConvertParameters,
    ) -> TransformationCommand:
        """
        Returns backend-specific convert insertion command.

        :param target_point: Target location for the correction.
        :param parameters: FakeConvertParameters to calculate activation quantization parameters.
        :return: Backend-specific TransformationCommand for the quantizer insertion operation.
        """

    @staticmethod
    @abstractmethod
    def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        """
        Returns a list of NNCFNodes to use as start nodes for activation path tracing.

        :param nncf_graph: NNCFGraph to get the start nodes.
        :return: List of NNCFNodes to use as start nodes for activation path tracing.

        """

    @staticmethod
    @abstractmethod
    def get_target_point_shape(nncf_graph: NNCFGraph, node: NNCFNode, target_point: TargetPoint) -> Tuple[int, ...]:
        """
        Returns shape of a target point tensor.

        :param nncf_graph: NNCFGraph instance.
        :param node: NNCFNode.
        :param target_point: Target point of which tensor shape is seeked.
        :return: Shape of target point tensor.
        """

    @staticmethod
    @abstractmethod
    def get_weight_quantization_axes(node: NNCFNode, target_point: TargetPoint, ndims: int) -> Tuple[int, ...]:
        """
        Returns axes for per-channel quantization of weights of the node placed on a input port_id.

        :param node: Quantized node with the weight.
        :param target_point: Corresponding target point.
        :param ndims: Number of dimensions of weight.
        :return: Axes for per-channel quantization of weights.
        """

    @staticmethod
    @abstractmethod
    def get_weight_tensor_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Optional[int]]:
        """
        Returns node's input port indices with weight tensors.

        :param node: NNCFNode to find its weight input port indices.
        :param graph: NNCFGraph instance.
        :return: Weights input port indices.
        """

    @staticmethod
    def get_weight_name(nncf_graph: NNCFGraph, target_point: TargetPoint) -> str:
        """
        Returns node's weight name corresponding to port ID.

        :param nncf_graph: NNCFGraph instance.
        :param target_point: The TargetPoint instance that contains layer's information.
        :return: Weight name.
        """

    @staticmethod
    def should_quantize_weight(weight_name: str, quantized_weight_names: Set[str]) -> bool:
        """
        Return True if weight should be quantized.

        :param weight_name: Weight name.
        :param quantized_weight_names: Set containing already quantized weight names.
        :return: A boolean value specifying whether a weight should be quantized.
        """

    @staticmethod
    @abstractmethod
    def get_ignored_metatypes(model_type: ModelType, device: TargetDevice) -> List[OperatorMetatype]:
        """
        Returns ignored metatypes based on a model type and device parameters.

        :param model_type: Model type parameter.
        :param device: Target device.
        :return: List of ignored metatypes.
        """

    @staticmethod
    @abstractmethod
    def get_ignored_names_by_layer_attributes(nncf_graph: NNCFGraph) -> Set[str]:
        """
        Returns names of ignored nodes based on layer_attributes.

        :param nncf_graph: NNCFGraph instance.
        :return: List of ignored names.
        """

    @abstractmethod
    def get_weight_nodes(self, nncf_graph: NNCFGraph) -> List[NNCFNode]:
        """
        Returns nodes that have weights.

        :param nncf_graph: Instance of NNCFGraph.
        :return: All nodes with weights.
        """

    @abstractmethod
    def is_matmul_with_constant(self, node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Returns true if given nncf matmul node is a matmul with a constant, False otherwise.

        :param Node: Instance of NNCFNode.
        :param nncf_graph: Instance of NNCFGraph.
        :return: True if given nncf matmul node is a matmul with a constant, False otherwise.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.quantization.quantizer_propagation.structs import QuantizationTrait
from nncf.tensorflow.graph.metatypes import common
from nncf.tensorflow.graph.metatypes import keras_layers as layer_metatypes
from nncf.tensorflow.graph.metatypes import tf_ops as op_metatypes
from nncf.tensorflow.graph.metatypes.common import LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_INPUTS
from nncf.tensorflow.graph.metatypes.common import LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_OUTPUTS
from nncf.tensorflow.graph.metatypes.common import LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_ONE_INPUT

# If a metatype is not in this list, then it is considered to be QuantizationTrait.NON_QUANTIZABLE.

DEFAULT_TF_QUANT_TRAIT_TO_OP_DICT = {
    QuantizationTrait.INPUTS_QUANTIZABLE: [
        *common.GENERAL_CONV_LAYER_METATYPES,
        *common.DEPTHWISE_CONV_LAYER_METATYPES,
        *common.DECONV_LAYER_METATYPES,
        *common.LINEAR_LAYER_METATYPES,
        *common.NORMALIZATION_LAYER_METATYPES,
        *common.ELEMENTWISE_LAYER_METATYPES,
        layer_metatypes.TFLocallyConnected1DLayerMetatype,
        layer_metatypes.TFLocallyConnected2DLayerMetatype,
        layer_metatypes.TFAveragePooling1DLayerMetatype,
        layer_metatypes.TFAveragePooling2DLayerMetatype,
        layer_metatypes.TFAveragePooling3DLayerMetatype,
        layer_metatypes.TFGlobalAveragePooling1DLayerMetatype,
        layer_metatypes.TFGlobalAveragePooling2DLayerMetatype,
        layer_metatypes.TFGlobalAveragePooling3DLayerMetatype,
        layer_metatypes.TFUpSampling1DLayerMetatype,
        layer_metatypes.TFUpSampling2DLayerMetatype,
        layer_metatypes.TFUpSampling3DLayerMetatype,
        layer_metatypes.TFAverageLayerMetatype,
        layer_metatypes.TFThresholdedReLULayerMetatype,
        layer_metatypes.TFELULayerMetatype,
        layer_metatypes.TFPReLULayerMetatype,
        layer_metatypes.TFLeakyReLULayerMetatype,
        layer_metatypes.TFActivationLayerMetatype,
        op_metatypes.TFAvgPoolOpMetatype,
        op_metatypes.TFAvgPool3DOpMetatype,
        op_metatypes.TFMeanOpMetatype,
        op_metatypes.TFResizeNearestNeighborOpMetatype,
        op_metatypes.TFEluOpMetatype,
        op_metatypes.TFLeakyReluOpMetatype,
        op_metatypes.TFRelu6OpMetatype,
        op_metatypes.TFBatchMatMulV2OpMetatype,
    ],
    QuantizationTrait.QUANTIZATION_AGNOSTIC: LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_ONE_INPUT
    + LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_INPUTS
    + LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_OUTPUTS,
    QuantizationTrait.CONCAT: [
        layer_metatypes.TFConcatenateLayerMetatype,
        op_metatypes.TFConcatOpMetatype,
    ],
    QuantizationTrait.OUTPUT_QUANTIZATION_AS_WEIGHTS: [layer_metatypes.TFEmbeddingLayerMetatype],
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path as osp

import tensorflow as tf
from tensorflow.keras.optimizers import schedules

from nncf.common.accuracy_aware_training.runner import BaseAccuracyAwareTrainingRunner
from nncf.common.accuracy_aware_training.runner import BaseAdaptiveCompressionLevelTrainingRunner
from nncf.common.logging import nncf_logger


class TFAccuracyAwareTrainingRunner(BaseAccuracyAwareTrainingRunner):
    """
    The Training Runner implementation for TensorFlow training code.
    """

    def validate(self, model):
        self.current_val_metric_value = self._validate_fn(model, epoch=self.cumulative_epoch_count)
        is_best = (not self.is_higher_metric_better) != (self.current_val_metric_value > self.best_val_metric_value)
        if is_best:
            self.best_val_metric_value = self.current_val_metric_value
        return self.current_val_metric_value

    def reset_training(self):
        self.configure_optimizers()

        optimizers = self.optimizer if isinstance(self.optimizer, (tuple, list)) else [self.optimizer]

        for optimizer in optimizers:
            scheduler = optimizer.learning_rate
            # pylint: disable=protected-access
            if isinstance(scheduler, tf.Variable) and not isinstance(
                optimizer._learning_rate, schedules.LearningRateSchedule
            ):
                scheduler = scheduler * self.base_lr_reduction_factor_during_search
                optimizer.learning_rate = scheduler
                optimizer.lr = scheduler
            elif isinstance(scheduler, (schedules.CosineDecay, schedules.ExponentialDecay)):
                scheduler.initial_learning_rate *= self.base_lr_reduction_factor_during_search
            elif isinstance(scheduler, schedules.PiecewiseConstantDecay):
                scheduler.values = [lr * self.base_lr_reduction_factor_during_search for lr in scheduler.values]
            else:
                nncf_logger.warning(
                    f"Learning rate scheduler {scheduler} is not supported yet. Won't change the learning rate."
                )

        self.training_epoch_count = 0
        self.best_val_metric_value = 0
        self.current_val_metric_value = 0

    def update_learning_rate(self):
        if self._update_learning_rate_fn is not None:
            self._update_learning_rate_fn(
                self.lr_scheduler, self.training_epoch_count, self.current_val_metric_value, self.current_loss
            )

    def _save_checkpoint(self, model, compression_controller, checkpoint_path):
        model.save_weights(checkpoint_path)

    def _load_checkpoint(self, model, checkpoint_path):
        if self._load_checkpoint_fn is not None:
            self._load_checkpoint_fn(model, checkpoint_path)
        else:
            model.load_weights(checkpoint_path)

    def _make_checkpoint_path(self, is_best, compression_rate=None):
        extension = ".pt"
        return osp.join(self._checkpoint_save_dir, f'acc_aware_checkpoint_{"best" if is_best else "last"}{extension}')

    def add_tensorboard_scalar(self, key, data, step):
        if self.verbose and self._tensorboard_writer is not None:
            self._tensorboard_writer.add_scalar(key, data, step)

    def add_tensorboard_image(self, key, data, step):
        if self.verbose and self._tensorboard_writer is not None:
            self._tensorboard_writer.add_image(key, data, step)


class TFAdaptiveCompressionLevelTrainingRunner(
    BaseAdaptiveCompressionLevelTrainingRunner, TFAccuracyAwareTrainingRunner
):
    def __init__(
        self,
        accuracy_aware_training_params,
        uncompressed_model_accuracy: float,
        verbose: bool = True,
        dump_checkpoints: bool = True,
        lr_updates_needed: bool = True,
        minimal_compression_rate: float = 0.0,
        maximal_compression_rate: float = 0.95,
    ):
        super().__init__(
            accuracy_aware_training_params,
            uncompressed_model_accuracy,
            verbose,
            dump_checkpoints,
            lr_updates_needed,
            minimal_compression_rate=minimal_compression_rate,
            maximal_compression_rate=maximal_compression_rate,
        )

    def _make_checkpoint_path(self, is_best, compression_rate=None):
        extension = ".pt"
        base_path = osp.join(self._checkpoint_save_dir, "acc_aware_checkpoint")
        if is_best:
            if compression_rate is None:
                raise ValueError("Compression rate cannot be None")
            return f"{base_path}_best_{compression_rate:.3f}{extension}"
        return f"{base_path}_last{extension}"



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import List


def should_consider_scope(scope_str: str, target_scopes: List[str], ignored_scopes: List[str]):
    # TODO: rewrite and add target_scopes handling
    return all(
        not re.fullmatch(ignored.replace("{re}", ""), scope_str) if ignored.startswith("{re}") else scope_str != ignored
        for ignored in ignored_scopes
    )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Optional, Tuple, TypeVar

import numpy as np

from nncf.common.graph.graph import NNCFNode
from nncf.parameters import CompressWeightsMode

TWeightType = TypeVar("TWeightType")


@dataclass
class WeightCompressionConfig:
    """
    Configuration on how to compress (quantize) a specific weight.

    :param mode: Defines a mode for weight compression. Defaults to INT8_ASYM mode.
    :param group_size: Number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
        The value -1 means no grouping. Defaults to -1.
    """

    mode: Optional[CompressWeightsMode] = CompressWeightsMode.INT8_ASYM
    group_size: Optional[int] = -1

    @property
    def num_bits(self):
        """
        :return: number of bits that is used for storing a single quantized value in the given mode.
        """
        return 8 if self.mode in [CompressWeightsMode.INT8_SYM, CompressWeightsMode.INT8_ASYM] else 4

    @property
    def is_asym_mode(self):
        return self.mode in [CompressWeightsMode.INT4_ASYM, CompressWeightsMode.INT8_ASYM]

    @property
    def is_integer(self):
        """
        :return: True if compression type in integer, else False.
        """
        return self.mode not in [CompressWeightsMode.NF4, CompressWeightsMode.E2M1]

    def __hash__(self):
        return hash((self.mode.value, self.group_size))

    def __str__(self):
        return f"{self.mode.value}_{self.group_size}"


@dataclass
class WeightCompressionParameters:
    """
    Weight compression parameters determine how and what weight should be compressed.

    :param weight_name: Unique weight name.
    :param node_with_weight: Node with weight in the NNCF graph.
    :param weight_port_id: Number of elements in the weight array.
    :param num_weights: Number of elements in the weight array.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param compression_config: Configuration of weight compression for the weight node.
    """

    weight_name: str
    node_with_weight: NNCFNode
    weight_port_id: int
    num_weights: np.uint64
    reduction_axes: Tuple[int, ...]
    compression_config: Optional[WeightCompressionConfig] = field(default_factory=WeightCompressionConfig)

    def __post_init__(self):
        # Explicitly cast num_weights to avoid overflow on finding total number of weights.
        # The issue happens on Windows, because np.ndarray.size() returns np.int32 and sum of weights is more than 2^32.
        self.num_weights = np.uint64(self.num_weights)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import networkx as nx  # type: ignore
import torch
from torch import nn

import nncf
import nncf.torch.graph.operator_metatypes as om
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import Dtype
from nncf.experimental.torch2.function_hook.graph.build_graph_mode import build_graph
from nncf.experimental.torch2.function_hook.graph.graph_utils import ConstMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import EdgeMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import FunctionMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import InOutMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import NodeType


def get_node_type(type: NodeType, meta: Union[ConstMeta, FunctionMeta, InOutMeta]) -> str:
    """
    Convert a given NodeType to its corresponding string representation.

    :param type: The type of the node, which can be one of the following:
    :param meta: Metadata associated with the node.
    :return: The string representation of the node type.
    """
    if isinstance(meta, InOutMeta) and type == NodeType.input:
        return "nncf_model_input"
    if isinstance(meta, InOutMeta) and type == NodeType.output:
        return "nncf_model_output"
    if isinstance(meta, ConstMeta):
        return "nncf_model_const"
    if isinstance(meta, FunctionMeta):
        return meta.fn_name
    raise nncf.InternalError("Unexpected metadata type")


def get_name_of_node(meta: Union[ConstMeta, FunctionMeta, InOutMeta]) -> str:
    """
    Get the name of a node based on its metadata.

    :param meta: The metadata of the node.
    :return: The name of the node.
    """
    if isinstance(meta, ConstMeta):
        return meta.name_in_model
    if isinstance(meta, FunctionMeta):
        return meta.op_name
    if isinstance(meta, InOutMeta):
        return meta.name
    raise nncf.InternalError("Unexpected metadata type")


def get_dtype(dtype: torch.dtype) -> Dtype:
    """
    Converts a torch dtype to a NNCF Dtype enum.

    :param dtype: The torch dtype to be converted.
    :return: The corresponding NNCF Dtype enum value.
    """
    if dtype in [torch.float, torch.float16, torch.bfloat16, torch.float32, torch.float64]:
        return Dtype.FLOAT
    return Dtype.INTEGER


def get_meta_type(node_type: str, meta: Union[ConstMeta, FunctionMeta, InOutMeta]) -> om.PTOperatorMetatype:
    """
    Converts the node type and metadata into a PTOperatorMetatype object.
    :param node_type: The type of the node.
    :param meta: The metadata associated with the node.
    :return: The PTOperatorMetatype object.
    """
    node_metatype = cast(om.PTOperatorMetatype, om.PT_OPERATOR_METATYPES.get_operator_metatype_by_op_name(node_type))
    node_sub_meta_type: Optional[om.PTOperatorMetatype] = None
    if node_metatype.get_subtypes() and isinstance(meta, FunctionMeta):
        node_sub_meta_type = node_metatype.determine_subtype(function_args=meta.args, functions_kwargs=meta.kwargs)
    return node_sub_meta_type or node_metatype


def convert_to_nncf_graph(nx_graph: nx.MultiDiGraph) -> NNCFGraph:
    """
    Converts a graph to an NNCFGraph.

    :param nx_graph: The graph to convert.
    :return: The converted NNCFGraph.
    """
    nncf_graph = NNCFGraph()

    map_nx_node_to_nncf_node: Dict[int, NNCFNode] = {}
    for node, data in nx_graph.nodes(data=True):
        meta: Union[ConstMeta, FunctionMeta, InOutMeta] = data["meta"]
        node_name = get_name_of_node(meta)
        node_type = get_node_type(data["type"], meta)
        meta_type = get_meta_type(node_type, meta)

        nncf_node = nncf_graph.add_nncf_node(
            node_name=node_name,
            node_type=node_type,
            node_metatype=meta_type,  # type: ignore[arg-type]
        )
        map_nx_node_to_nncf_node[node] = nncf_node

    map_edges: Dict[Tuple[int, int], List[EdgeMeta]] = defaultdict(list)

    for s_node, t_node, data in nx_graph.edges(data=True):
        meta = data["meta"]
        if isinstance(meta, EdgeMeta):
            map_edges[(s_node, t_node)].append(meta)

    for (s_node, t_node), list_meta in map_edges.items():
        source_node = map_nx_node_to_nncf_node[s_node]
        target_node = map_nx_node_to_nncf_node[t_node]
        nncf_graph.add_edge_between_nncf_nodes(
            source_node.node_id,
            target_node.node_id,
            tensor_shape=list_meta[0].shape,
            input_port_id=list_meta[0].input_port,
            output_port_id=list_meta[0].output_port,
            dtype=get_dtype(list_meta[0].dtype),
            parallel_input_port_ids=[meta.input_port for meta in list_meta[1:]] if len(list_meta) > 1 else None,
        )
    return nncf_graph


def build_nncf_graph(model: nn.Module, *args: Any, **kwargs: Any) -> NNCFGraph:
    """
    Builds an NNCF graph from the given PyTorch model.

    :param model: The PyTorch model to build the graph from.
    :param *args: Positional arguments to model inference.
    :param **kwargs: Keyword arguments to model inference.
    :return: The NNCF graph representation of the model.
    """
    graph = build_graph(model, *args, **kwargs)
    return convert_to_nncf_graph(graph)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Any, List, Optional, TypeVar

from nncf.common.engine import Engine
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype

TModel = TypeVar("TModel")
TPModel = TypeVar("TPModel")


class PreparedModel(ABC):
    @property
    @abstractmethod
    def model_for_inference(self) -> TPModel:
        """
        Returns prepared model for inference.

        :return: Prepared model for inference.
        """

    @property
    @abstractmethod
    def engine(self) -> Engine:
        """
        Returns the engine for inference the prepared model.

        :return: The engine for inference the prepared model.
        """

    def __call__(self, input_data: Any) -> Any:
        """
        Runs model on the provided input data and returns the raw model outputs.

        :param input_data: inputs for the model
        :return: raw model outputs
        """
        return self.engine.infer(input_data)


class AccuracyControlAlgoBackend(ABC):
    # Metatypes

    @staticmethod
    @abstractmethod
    def get_op_with_weights_metatypes() -> List[OperatorMetatype]:
        """
        Returns a list of operation metatypes that can be reverted to representation
        with int8 weights.

        :return: The list of operation metatypes that can be reverted to representation
            with int8 weights.
        """

    @staticmethod
    @abstractmethod
    def get_quantizer_metatypes() -> List[OperatorMetatype]:
        """
        Returns a list of quantizer metatypes.

        :return: The list of quantizer metatypes.
        """

    @staticmethod
    @abstractmethod
    def get_const_metatypes() -> List[OperatorMetatype]:
        """
        Returns a list of constant metatypes.

        :return: The list of constant metatypes.
        """

    @staticmethod
    @abstractmethod
    def get_quantizable_metatypes() -> List[OperatorMetatype]:
        """
        Returns a list of metatypes for operations that may be quantized.

        :return: The list of metatypes for operations that may be quantized.
        """

    @staticmethod
    @abstractmethod
    def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        """
        Returns a list of NNCFNodes to use as start nodes for activation path tracing.

        :param nncf_graph: NNCFGraph to get the start nodes.
        :return: List of NNCFNodes to use as start nodes for activation path tracing.
        """

    @staticmethod
    @abstractmethod
    def get_quantize_agnostic_metatypes() -> List[OperatorMetatype]:
        """
        Returns a list of quantize agnostic metatypes.

        :return: The list of quantize agnostic metatypes.
        """

    @staticmethod
    @abstractmethod
    def get_shapeof_metatypes() -> List[OperatorMetatype]:
        """
        Returns a list of shape of metatypes.

        :return: The list of shape of metatypes.
        """

    # Manipulations with bias value and weights

    @staticmethod
    @abstractmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        """
        Checks if the node has a bias or not.

        :param node: The node to check.
        :param nncf_graph: The NNCF graph.
        :return: True` if `node` corresponds to the operation with bias
            (bias is added to the output tensor of that operation), `False` otherwise.
        """

    @staticmethod
    @abstractmethod
    def is_node_with_weight(node: NNCFNode) -> bool:
        """
        Checks if the node has a weight or not.

        :param node: The node to check.
        :param nncf_graph: The NNCF graph.
        :return: `True` if `node` corresponds to the operation with weights, `False` otherwise.
        """

    @staticmethod
    @abstractmethod
    def get_bias_value(node_with_bias: NNCFNode, nncf_graph: NNCFGraph, model: TModel) -> Any:
        """
        Returns the bias value for the biased node.

        :param node_with_bias: The node that corresponds to the operation with bias.
        :param nncf_graph: The NNCF graph.
        :param model: The model that contains this operation.
        :return: The bias value that is applied to the output tensor of the node's operation.
        """

    @staticmethod
    @abstractmethod
    def get_weight_value(node_with_weight: NNCFNode, model: TModel, port_id: int) -> Any:
        """
        Returns the weight value for the node with weight.

        :param node_with_weight: The node with weight.
        :param model: The model that contains this operation.
        :param port_id: The input port ID to get weight input.
        :return: The weight value.
        """

    @staticmethod
    @abstractmethod
    def get_weight_tensor_port_ids(node: NNCFNode) -> List[Optional[int]]:
        """
        Returns node's input port indices with weights tensors.

        :param node: NNCFNode to find its weights input port indices.
        :return: Weights input port indices.
        """

    @staticmethod
    @abstractmethod
    def get_model_size(model: TModel) -> int:
        """
        Returns model size

        :param model: A model
        :return: Model size (in bytes)
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import networkx as nx

import nncf
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.pruning.utils import get_input_masks
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_width import ElasticWidthHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.multi_elasticity_handler import MultiElasticityHandler
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import PTModuleConv2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv2dSubtype


class SubnetGraph:
    """
    Graph that represents active subnet in convenient way for visualization.
    """

    def __init__(self, compression_graph: PTNNCFGraph, multi_elasticity_handler: MultiElasticityHandler):
        # TODO: visualize other elastic dimension: depth, kernel (ticket 76870)
        self._width_graph = compression_graph.get_graph_for_structure_analysis(extended=True)
        for node_key, compression_node in compression_graph.nodes.items():
            operator_name = self._get_operator_name(compression_node, multi_elasticity_handler.width_handler)

            metatype = compression_node.metatype
            color = None
            if metatype == PTModuleConv2dMetatype:
                color = "lightblue"
            if metatype == PTModuleDepthwiseConv2dSubtype:
                operator_name = f"DW_{operator_name}"
                color = "purple"

            target_node_to_draw = self._width_graph.nodes[node_key]
            target_node_to_draw["label"] = operator_name
            target_node_to_draw["style"] = "filled"
            if color is not None:
                target_node_to_draw["color"] = color

    def get(self) -> nx.DiGraph:
        return self._width_graph

    @staticmethod
    def _get_operator_name(compressed_node: NNCFNode, width_handler: ElasticWidthHandler):
        operator_name = compressed_node.node_type
        node_id = compressed_node.node_id
        node = SubnetGraph._get_original_node_from_compressed_node_name(compressed_node.node_name, width_handler)
        if node is not None:
            operator_name = node.node_type
            node_id = node.node_id
            input_masks = get_input_masks(node, width_handler.propagation_graph)
            input_widths = None
            if input_masks:
                input_widths = [ElasticWidthHandler.mask_to_width(input_mask) for input_mask in input_masks]
            output_width = ElasticWidthHandler.mask_to_width(node.attributes["output_mask"])

            if input_widths:
                IW = None
                if len(input_widths) == 1 and input_widths[0]:
                    IW = input_widths[0]
                if len(input_widths) > 1 and all(input_widths):
                    IW = input_widths
                if IW is not None:
                    operator_name += f"_IW{IW}"

            if output_width is not None:
                operator_name += f"_OW{output_width}"

            group_id = width_handler.get_group_id_by_node_name(node.node_name)
            if group_id is not None:
                operator_name += f"_G{group_id}"

        operator_name += f"_#{node_id}"
        return operator_name

    @staticmethod
    def _get_original_node_from_compressed_node_name(
        node_name: NNCFNodeName, width_handler: ElasticWidthHandler
    ) -> Optional[NNCFNode]:
        try:
            propagation_graph: PTNNCFGraph = width_handler.propagation_graph
            result = propagation_graph.get_node_by_name(node_name)
        except nncf.InternalError:
            result = None
        return result



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

import jsonschema
import jstyleson as json  # type: ignore

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api
from nncf.common.utils.os import safe_open
from nncf.config.definitions import SCHEMA_VISUALIZATION_URL
from nncf.config.schema import NNCF_CONFIG_SCHEMA
from nncf.config.schema import REF_VS_ALGO_SCHEMA
from nncf.config.schema import validate_single_compression_algo_schema
from nncf.config.structures import NNCFExtraConfigStruct


@api(canonical_alias="nncf.NNCFConfig")
class NNCFConfig(dict[str, Any]):
    """Contains the configuration parameters required for NNCF to apply the selected algorithms.

    This is a regular dictionary object extended with some utility functions, such as the ability to attach well-defined
    structures to pass non-serializable objects as parameters. It is primarily built from a .json file, or from a
    Python JSON-like dictionary - both data types will be checked against a JSONSchema. See the definition of the
    schema at https://openvinotoolkit.github.io/nncf/schema/, or by calling NNCFConfig.schema()."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__nncf_extra_structs: Dict[str, NNCFExtraConfigStruct] = {}

    @classmethod
    def from_dict(cls, nncf_dict: Dict[str, Any]) -> "NNCFConfig":
        """
        Load NNCF config from a Python dictionary. The dict must contain only JSON-supported primitives.

        :param nncf_dict: A Python dict with the JSON-style configuration for NNCF.
        """

        cls.validate(nncf_dict)
        return cls(deepcopy(nncf_dict))

    @classmethod
    def from_json(cls, path: str) -> "NNCFConfig":
        """
        Load NNCF config from a JSON file at `path`.

        :param path: Path to the .json file containing the NNCF configuration.
        """
        file_path = Path(path)
        with safe_open(file_path) as f:
            loaded_json = json.load(f)
        return cls.from_dict(loaded_json)

    def register_extra_structs(self, struct_list: List[NNCFExtraConfigStruct]) -> None:
        """
        Attach the supplied list of extra configuration structures to this configuration object.

        :param struct_list: List of extra configuration structures.
        """
        for struct in struct_list:
            struct_id = struct.get_id()
            if struct_id in self.__nncf_extra_structs:
                raise nncf.InternalError(f"{struct_id} is already registered as extra struct in NNCFConfig!")
            self.__nncf_extra_structs[struct_id] = struct

    def get_extra_struct(self, struct_cls: Type[NNCFExtraConfigStruct]) -> NNCFExtraConfigStruct:
        return self.__nncf_extra_structs[struct_cls.get_id()]

    def has_extra_struct(self, struct_cls: Type[NNCFExtraConfigStruct]) -> bool:
        return struct_cls.get_id() in self.__nncf_extra_structs

    def get_all_extra_structs(self) -> List[NNCFExtraConfigStruct]:
        return list(self.__nncf_extra_structs.values())

    def get_redefinable_global_param_value_for_algo(self, param_name: str, algo_name: str) -> Optional[str]:
        """
        Some parameters can be specified both on the global NNCF config .json level (so that they apply
        to all algos), and at the same time overridden in the algorithm-specific section of the .json.
        This function returns the value that should apply for a given algorithm name, considering the
        exact format of this config.

        :param param_name: The name of a parameter in the .json specification of the NNCFConfig, that may
          be present either at the top-most level of the .json, or at the top level of the algorithm-specific
          subdict.
        :param algo_name: The name of the algorithm (among the allowed algorithm names in the .json) for which
          the resolution of the redefinable parameter should occur.
        :return: The value of the parameter that should be applied for the algo specified by `algo_name`.
        """
        from nncf.config.extractors import extract_algo_specific_config

        algo_config = extract_algo_specific_config(self, algo_name)
        param = self.get(param_name)
        algo_specific_param = algo_config.get(param_name)
        if algo_specific_param is not None:
            param = algo_specific_param
        return param

    @staticmethod
    def schema() -> Dict[str, Any]:
        """
        Returns the JSONSchema against which the input data formats (.json or Python dict) are validated.
        """
        return NNCF_CONFIG_SCHEMA

    @staticmethod
    def _is_path_to_algorithm_name(path_parts: List[str]) -> bool:
        return (len(path_parts) == 2 and path_parts[0] == "compression" and path_parts[1] == "algorithm") or (
            len(path_parts) == 3
            and path_parts[0] == "compression"
            and path_parts[1].isnumeric()
            and path_parts[2] == "algorithm"
        )

    @staticmethod
    def validate(loaded_json: Dict[str, Any]) -> None:
        try:
            jsonschema.validate(loaded_json, NNCFConfig.schema())
        except jsonschema.ValidationError as e:
            nncf_logger.error("Invalid NNCF config supplied!")
            absolute_path_parts = [str(x) for x in e.absolute_path]
            if not NNCFConfig._is_path_to_algorithm_name(absolute_path_parts):
                e.message += f"\nRefer to the NNCF config schema documentation at {SCHEMA_VISUALIZATION_URL}"
                e.schema = "*schema too long for stdout display*"  # type: ignore[assignment]
                raise e

            # Need to make the error more algo-specific in case the config was so bad that no
            # scheme could be matched
            # If error is in the algo section, will revalidate the algo sections separately to
            # make the error message more targeted instead of displaying the entire huge schema.
            compression_section = loaded_json["compression"]
            if isinstance(compression_section, dict):
                validate_single_compression_algo_schema(compression_section, REF_VS_ALGO_SCHEMA)
            else:
                # Passed a list of dicts
                for compression_algo_dict in compression_section:
                    validate_single_compression_algo_schema(compression_algo_dict, REF_VS_ALGO_SCHEMA)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path as osp
from typing import Any, Dict, List, Optional, Tuple

from nncf.common.compression import BaseCompressionAlgorithmController as BaseController
from nncf.common.utils.debug import set_debug_log_dir
from nncf.config import NNCFConfig
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.model_creation import create_compression_algorithm_builder_from_algo_names
from nncf.torch.model_creation import synchronize_all_processes_in_distributed_mode
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.utils import is_main_process


def create_compressed_model_from_algo_names(
    nncf_network: NNCFNetwork, config: NNCFConfig, algo_names: List[str], dump_graphs: bool = True
) -> Tuple[BaseController, NNCFNetwork]:
    """
    The main function used to produce a model ready for compression fine-tuning from empty NNCFNetwork,
    a configuration object and a list of compression algorithm names.

    :param nncf_network: empty NNCFNetwork, that just wrapped original PyTorch model.
    :param config: A configuration object used to determine the exact compression modifications to be applied
    to the model
    :param algo_names: list of compression algorithm names.
    :param dump_graphs: Whether or not should also dump the internal graph representation of the
    original and compressed models in the .dot format into the log directory.
    :return: A controller for the compression algorithm (or algorithms, in which case the controller
    is an instance of CompositeCompressionController) and the model ready for compression parameter training wrapped
    as an object of NNCFNetwork.
    """
    set_debug_log_dir(config.get("log_dir", ""))

    if dump_graphs:
        original_model_graph = nncf_network.nncf.get_original_graph()
        original_model_graph.visualize_graph(osp.join(config.get("log_dir", ""), "original_graph.dot"))

    builder = create_compression_algorithm_builder_from_algo_names(algo_names, config, should_init=True)

    compressed_model = builder.apply_to(nncf_network)
    compression_ctrl = builder.build_controller(compressed_model)

    # Required to ensure that the model leaving create_compressed_model has correct compressed graph.
    # In particular, this is currently required for correct functioning of RNNs.
    compressed_model.nncf.rebuild_graph()

    if dump_graphs and is_main_process():
        original_model_graph = compressed_model.nncf.get_graph()
        original_model_graph.visualize_graph(osp.join(config.get("log_dir", ""), "compressed_graph.dot"))

    synchronize_all_processes_in_distributed_mode()
    return compression_ctrl, compressed_model


def resume_compression_algorithm_builder(
    compression_state: Dict[str, Any], config: Optional[NNCFConfig] = None
) -> PTCompressionAlgorithmBuilder:
    """
    Resume compression builder from its state.

    :param compression_state: representation of the entire compression state to unambiguously restore
    the compressed model. Includes builder and controller states.
    :param config: A configuration object used to determine the exact compression modifications to be applied
    to the model. NNCFConfig is required for resume from checkpoint, because currently CompressionBuilder takes
    NNCFConfig to get some params, like BNAdaptInitArgs. It's needed not only on init. e.g.
    BNAdaptInitArgs.train_loader is used by ProgressiveShrinkingAlgorithm during the training.
    :return: resumed compression builder
    """
    if config is None:
        config = NNCFConfig()

    builder_state = compression_state[BaseController.BUILDER_STATE]
    algo_names = list(builder_state)
    builder = create_compression_algorithm_builder_from_algo_names(algo_names, config, should_init=False)
    builder.load_state(builder_state)
    return builder


def resume_compression_from_state(
    nncf_network: NNCFNetwork, compression_state: Dict[str, Any], config: Optional[NNCFConfig] = None
) -> Tuple[NNCFNetwork, PTCompressionAlgorithmController]:
    """
    Resumes compression model

    :param nncf_network: empty NNCFNetwork, that just wrapped original PyTorch model.
    :param compression_state: representation of the entire compression state to unambiguously restore
    the compressed model. Includes builder and controller states.
    :param config: is needed for overriding state or for passing extra structs only, like BNAdaptInitArgs required
    during training, like in NAS.
    :return: the model ready for compression parameter training wrapped as an object of NNCFNetwork and
    a controller for the compression algorithm.
    """
    builder = resume_compression_algorithm_builder(compression_state, config)
    model = builder.apply_to(nncf_network)
    ctrl = builder.build_controller(nncf_network)
    ctrl.load_state(compression_state[BaseController.CONTROLLER_STATE])
    return model, ctrl



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.utils import get_first_nodes_of_type
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import ActivationQuantizationInsertionPoint
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.stateful_classes_registry import TF_STATEFUL_CLASSES
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.backend import copy_model
from nncf.config.extractors import extract_range_init_params
from nncf.experimental.tensorflow.graph.converter import SubclassedConverter
from nncf.experimental.tensorflow.graph.model_transformer import TFModelTransformerV2
from nncf.experimental.tensorflow.graph.transformations.commands import TFTargetPoint
from nncf.experimental.tensorflow.graph.transformations.layout import TFTransformationLayoutV2
from nncf.experimental.tensorflow.nncf_network import NNCFNetwork
from nncf.experimental.tensorflow.quantization.init_range import RangeInitializerV2
from nncf.experimental.tensorflow.quantization.init_range import TFRangeInitParamsV2
from nncf.experimental.tensorflow.quantization.quantizers import create_quantizer
from nncf.tensorflow.algorithm_selector import TF_COMPRESSION_ALGORITHMS
from nncf.tensorflow.graph.metatypes.tf_ops import TFOpWithWeightsMetatype
from nncf.tensorflow.graph.transformations.commands import TFInsertionCommand
from nncf.tensorflow.quantization.algorithm import QuantizationBuilder
from nncf.tensorflow.quantization.algorithm import QuantizationController
from nncf.tensorflow.quantization.algorithm import TFQuantizationPoint
from nncf.tensorflow.quantization.algorithm import TFQuantizationPointStateNames
from nncf.tensorflow.quantization.algorithm import TFQuantizationSetup
from nncf.tensorflow.quantization.quantizers import TFQuantizerSpec

UNSUPPORTED_TF_OP_METATYPES = []


class TFQuantizationPointV2StateNames(TFQuantizationPointStateNames):
    IS_WEIGHT_QUANTIZATION = "is_weight_quantization"
    INPUT_SHAPE = "input_shape"
    CHANNEL_AXES = "channel_axes"


class TFQuantizationPointV2(TFQuantizationPoint):
    _state_names = TFQuantizationPointV2StateNames

    def __init__(
        self,
        op_name: str,
        quantizer_spec: TFQuantizerSpec,
        target_point: TargetPoint,
        is_weight_quantization: bool,
        input_shape: Optional[List[int]] = None,
        channel_axes: Optional[List[int]] = None,
    ):
        super().__init__(op_name, quantizer_spec, target_point)
        self.is_weight_quantization = is_weight_quantization
        self.input_shape = input_shape
        self.channel_axes = channel_axes

    def get_state(self) -> Dict[str, Any]:
        state = super().get_state()
        state.update(
            {
                self._state_names.IS_WEIGHT_QUANTIZATION: self.is_weight_quantization,
                self._state_names.INPUT_SHAPE: self.input_shape,
                self._state_names.CHANNEL_AXES: self.channel_axes,
            }
        )
        return state

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFQuantizationPointV2":
        target_point_cls = TF_STATEFUL_CLASSES.get_registered_class(state[cls._state_names.TARGET_POINT_CLASS_NAME])
        kwargs = {
            cls._state_names.OP_NAME: state[cls._state_names.OP_NAME],
            cls._state_names.QUANTIZER_SPEC: TFQuantizerSpec.from_state(state[cls._state_names.QUANTIZER_SPEC]),
            cls._state_names.TARGET_POINT: target_point_cls.from_state(state[cls._state_names.TARGET_POINT]),
            cls._state_names.IS_WEIGHT_QUANTIZATION: state[cls._state_names.IS_WEIGHT_QUANTIZATION],
            cls._state_names.INPUT_SHAPE: state[cls._state_names.INPUT_SHAPE],
            cls._state_names.CHANNEL_AXES: state[cls._state_names.CHANNEL_AXES],
        }
        return cls(**kwargs)


class TFQuantizationSetupV2(TFQuantizationSetup):
    @classmethod
    def from_state(cls, state: Dict) -> "TFQuantizationSetupV2":
        setup = TFQuantizationSetupV2()
        for quantization_point_state in state[cls._state_names.QUANTIZATION_POINTS]:
            quantization_point = TFQuantizationPointV2.from_state(quantization_point_state)
            setup.add_quantization_point(quantization_point)

        if cls._state_names.UNIFIED_SCALE_GROUPS in state:
            for quantization_group in state[cls._state_names.UNIFIED_SCALE_GROUPS]:
                setup.register_unified_scale_group(quantization_group)
        return setup


def _get_quantizer_op_name(prefix: str, is_wq: bool, port_id: int, target_type) -> str:
    pos = "pre_hook" if target_type == TargetType.OPERATOR_PRE_HOOK else "post_hook"
    qtype = "W" if is_wq else "A"
    name = "_".join([pos, qtype, str(port_id)])
    quantizer_op_name = f"{prefix}/{name}"
    return quantizer_op_name


def _get_tensor_specs(
    node: NNCFNode, nncf_graph: NNCFGraph, port_ids: List[int], is_input_tensors: bool, is_weight_tensors: bool
):
    """
    Returns specification of tensors for `node` according to `port_ids`.
    """
    tensor_specs = []

    if is_weight_tensors:
        assert is_input_tensors
        metatype = node.metatype

        assert len(port_ids) == 1
        assert len(metatype.weight_definitions) == 1

        channel_axes = metatype.weight_definitions[0].channel_axes
        weight_shape = node.layer_attributes.get_weight_shape()
        tensor_specs.append((weight_shape, channel_axes))
    else:
        data_format = node.layer_attributes.get_data_format()
        channel_axes = [-1] if data_format == "channels_last" else [1]

        if is_input_tensors:
            edges = nncf_graph.get_input_edges(node)
            for input_port_id in port_ids:
                tensor_specs.extend((e.tensor_shape, channel_axes) for e in edges if e.input_port_id == input_port_id)
        else:
            edges = nncf_graph.get_output_edges(node)
            for output_port_id in port_ids:
                filtered_edges = [e for e in edges if e.output_port_id == output_port_id]

                shape = filtered_edges[0].tensor_shape
                for e in filtered_edges:
                    assert e.tensor_shape == shape

                tensor_specs.append((shape, channel_axes))

    assert len(tensor_specs) == len(port_ids)

    return tensor_specs


@TF_COMPRESSION_ALGORITHMS.register("experimental_quantization")
class QuantizationBuilderV2(QuantizationBuilder):
    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        quantizer_setup_state = state_without_name[self._state_names.QUANTIZER_SETUP]
        self._quantizer_setup = TFQuantizationSetupV2.from_state(quantizer_setup_state)

    def _parse_range_init_params(self) -> TFRangeInitParamsV2:
        range_init_params = extract_range_init_params(self.config, self.name)
        return TFRangeInitParamsV2(**range_init_params) if range_init_params is not None else None

    def _build_insertion_commands_for_quantizer_setup(
        self, quantizer_setup: TFQuantizationSetupV2
    ) -> List[TFInsertionCommand]:
        insertion_commands = []
        quantization_points = quantizer_setup.get_quantization_points()
        # quantization point id is her index inside the `quantization_points` list
        was_processed = {qp_id: False for qp_id in range(len(quantization_points))}

        for unified_scales_group in quantizer_setup.get_unified_scale_groups():
            qp = quantization_points[unified_scales_group[0]]
            quantizer = create_quantizer(
                f"{qp.op_name}/unified_scale_group",
                qp.quantizer_spec,
                qp.is_weight_quantization,
                qp.input_shape,
                qp.channel_axes,
            )

            self._op_names.append(quantizer.name)

            for qp_id in unified_scales_group:
                if was_processed[qp_id]:
                    raise nncf.InternalError("Unexpected behavior")
                was_processed[qp_id] = True

                curr_qp = quantization_points[qp_id]
                # Checks
                assert curr_qp.quantizer_spec.get_state() == qp.quantizer_spec.get_state()
                assert curr_qp.input_shape == qp.input_shape
                assert curr_qp.channel_axes == qp.channel_axes

                command = TFInsertionCommand(
                    curr_qp.target_point, quantizer, TransformationPriority.QUANTIZATION_PRIORITY
                )
                insertion_commands.append(command)

        for qp_id, qp in enumerate(quantization_points):
            if was_processed[qp_id]:
                continue

            quantizer = create_quantizer(
                qp.op_name, qp.quantizer_spec, qp.is_weight_quantization, qp.input_shape, qp.channel_axes
            )

            self._op_names.append(quantizer.name)

            command = TFInsertionCommand(qp.target_point, quantizer, TransformationPriority.QUANTIZATION_PRIORITY)
            insertion_commands.append(command)

        return insertion_commands

    def get_transformation_layout(self, model: NNCFNetwork):
        transformations = TFTransformationLayoutV2()
        if self._quantizer_setup is None:
            self._quantizer_setup = self._get_quantizer_setup(model)
        insertion_commands = self._build_insertion_commands_for_quantizer_setup(self._quantizer_setup)
        for command in insertion_commands:
            transformations.register(command)
        return transformations

    def _build_controller(self, model: NNCFNetwork) -> "QuantizationControllerV2":
        return QuantizationControllerV2(model, self.config, self._op_names)

    def _run_range_initialization(self, model: NNCFNetwork) -> None:
        if self._range_initializer is None:
            self._range_initializer = RangeInitializerV2(self._range_init_params)
        self._range_initializer.run(model)

    def _get_input_preprocessing_nodes(self, nncf_graph: NNCFGraph, model: NNCFNetwork) -> List[NNCFNode]:
        return []

    def _get_quantizer_setup(self, model: NNCFNetwork) -> TFQuantizationSetupV2:
        converter = SubclassedConverter(model, model.input_signature)
        nncf_graph = converter.convert()
        # Find out which metatypes unsupported by the quantization algorithm
        for node in nncf_graph.get_all_nodes():
            if node.metatype in UNSUPPORTED_TF_OP_METATYPES:
                nncf_logger.warning(
                    "The operation {} is unsupported by the quantization algorithm.".format(node.node_name)
                )

        # Possible configurations of quantizer for nodes with weights.
        possible_qconfigs_for_nodes_with_weight = self._get_quantizable_weighted_layer_nodes(nncf_graph)
        qp_solution = self._get_quantizer_propagation_solution(
            nncf_graph, possible_qconfigs_for_nodes_with_weight, [], model
        )

        # Logic of the TFQuantizationSetupV2 creation

        quantization_setup = TFQuantizationSetupV2()
        node_name_to_qconfig_map: Dict[str, QuantizerConfig] = {}
        qp_id_to_setup_index_map: Dict[QuantizationPointId, int] = {}
        first_conv_nodes = get_first_nodes_of_type(nncf_graph, ["Conv2D", "Conv3D"])

        for idx, (qp_id, qp) in enumerate(qp_solution.quantization_points.items()):
            qp_id_to_setup_index_map[qp_id] = idx
            target_node = nncf_graph.get_node_by_name(qp.insertion_point.target_node_name)

            if qp.is_weight_quantization_point():
                # Check correctness
                if target_node.node_name in node_name_to_qconfig_map:
                    assigned_qconfig = node_name_to_qconfig_map[target_node.node_name]
                    if qp.qconfig != assigned_qconfig:
                        raise nncf.InternalError(
                            "Inconsistent quantizer configurations selected by solver for one "
                            f"and the same quantizable op! Tried to assign {qp.qconfig} to "
                            f"{target_node.node_name} as specified by QP {qp_id}, but the op "
                            f"already has quantizer config {assigned_qconfig} assigned to it!"
                        )
                    continue  # The operation has already been quantized
                node_name_to_qconfig_map[target_node.node_name] = qp.qconfig

                # Parameters
                half_range = self._get_half_range(qp.qconfig, target_node, first_conv_nodes)
                narrow_range = not half_range
                target_type = TargetType.OPERATOR_PRE_HOOK
                if not issubclass(target_node.metatype, TFOpWithWeightsMetatype):
                    raise nncf.InternalError(f"Unexpected type of metatype: {type(target_node.metatype)}")
                port_ids = [weight_def.port_id for weight_def in target_node.metatype.weight_definitions]

            else:
                assert qp.is_activation_quantization_point()

                # Check correctness
                if not isinstance(qp.insertion_point, ActivationQuantizationInsertionPoint):
                    raise nncf.InternalError(f"Unexpected type of insertion point: {type(qp.insertion_point)}")

                # Parameters
                half_range = False
                narrow_range = False
                if qp.insertion_point.input_port_id is not None:
                    port_ids = [qp.insertion_point.input_port_id]  # Input port ids
                    target_type = TargetType.OPERATOR_PRE_HOOK
                else:
                    port_ids = [0]  # Output port ids
                    target_type = TargetType.OPERATOR_POST_HOOK

            tensor_specs = _get_tensor_specs(
                target_node,
                nncf_graph,
                port_ids,
                target_type == TargetType.OPERATOR_PRE_HOOK,
                qp.is_weight_quantization_point(),
            )

            for port_id, (tensor_shape, channel_axes) in zip(port_ids, tensor_specs):
                quantizer_op_name = _get_quantizer_op_name(
                    target_node.node_name, qp.is_weight_quantization_point(), port_id, target_type
                )
                quantizer_spec = TFQuantizerSpec.from_config(qp.qconfig, narrow_range, half_range)
                target_point = TFTargetPoint(target_node.node_name, target_node.node_type, port_id, target_type)
                qpoint = TFQuantizationPointV2(
                    quantizer_op_name,
                    quantizer_spec,
                    target_point,
                    qp.is_weight_quantization_point(),
                    tensor_shape,
                    channel_axes,
                )

                quantization_setup.add_quantization_point(qpoint)

        # Registration of unified scale groups
        for unified_group in qp_solution.unified_scale_groups.values():
            us_group = [qp_id_to_setup_index_map[qp_id] for qp_id in unified_group]
            quantization_setup.register_unified_scale_group(us_group)

        return quantization_setup

    def apply_to(self, model: NNCFNetwork) -> NNCFNetwork:
        transformation_layout = self.get_transformation_layout(model)
        transformer = TFModelTransformerV2(model)
        transformed_model = transformer.transform(transformation_layout)

        if self.should_init:
            self.initialize(transformed_model)

        return transformed_model


class QuantizationControllerV2(QuantizationController):
    def strip_model(self, model: NNCFNetwork, do_copy: bool = False) -> NNCFNetwork:
        if do_copy:
            model = copy_model(model)
        return model

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        return NNCFStatistics()

    def prepare_for_export(self) -> None:
        self._model.compute_output_shape(self._model.input_signature.shape.as_list())
        self._model = self.strip_model(self._model)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.quantization.quantizer_propagation.structs import QuantizationTrait
from nncf.onnx.graph.metatypes import onnx_metatypes
from nncf.onnx.graph.metatypes.groups import INPUTS_QUANTIZABLE_OPERATIONS
from nncf.onnx.graph.metatypes.groups import QUANTIZE_AGNOSTIC_OPERATIONS

# If a metatype is not in this list, then it is considered to be QuantizationTrait.NON_QUANTIZABLE.

DEFAULT_ONNX_QUANT_TRAIT_TO_OP_DICT = {
    QuantizationTrait.INPUTS_QUANTIZABLE: INPUTS_QUANTIZABLE_OPERATIONS,
    QuantizationTrait.QUANTIZATION_AGNOSTIC: QUANTIZE_AGNOSTIC_OPERATIONS,
    QuantizationTrait.CONCAT: [onnx_metatypes.ONNXConcatMetatype],
    QuantizationTrait.OUTPUT_QUANTIZATION_AS_WEIGHTS: [onnx_metatypes.ONNXEmbeddingMetatype],
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Set

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.scopes import IgnoredScope
from nncf.scopes import get_difference_ignored_scope
from nncf.scopes import get_matched_ignored_scope_info


@dataclass
class TargetScope(IgnoredScope):
    """
    Specifies the target portions in a model graph.

    Example:

    ..  code-block:: python
        # Specified by node names:
        node_names = ['node_1', 'node_2', 'node_3']
        target_scope = TargetScope(names=node_names)

        # Specified using regular expressions:
        patterns = ['.*node_\\d']
        target_scope = TargetScope(patterns=patterns)

        # Specified by operation types, e.g.,

        # OpenVINO opset https://docs.openvino.ai/latest/openvino_docs_ops_opset.html
        operation_types = ['Multiply', 'GroupConvolution', 'Interpolate']
        target_scope = TargetScope(types=operation_types)

        # ONNX opset https://github.com/onnx/onnx/blob/main/docs/Operators.md
        operation_types = ['Mul', 'Conv', 'Resize']
        target_scope = TargetScope(types=operation_types)

        # Specifies by subgraphs:
        from nncf import Subgraph
        target_scope = TargetScope(subgraphs=[
            Subgraph(inputs=["node_1"], outputs=["node_3"])
        ])

    **Note:** Operation types must be specified according to the model framework.

    :param names: List of target node names.
    :param patterns: List of regular expressions that define patterns for names of target nodes.
    :param types: List of target operation types.
    :param subgraphs: List of target subgraphs.
    :param validate: If set to True, then a RuntimeError will be raised if any target scope does not match
      in the model graph.
    """

    def __hash__(self) -> int:
        return hash(
            (
                frozenset(self.names),
                frozenset(self.patterns),
                frozenset(self.types),
                frozenset((frozenset(subgraph.inputs), frozenset(subgraph.outputs)) for subgraph in self.subgraphs),
                self.validate,
            )
        )


def get_target_node_names_from_target_scope(
    target_scope: TargetScope, nncf_graph: NNCFGraph, strict: bool = True
) -> Set[str]:
    """
    Returns NNCF node names from the graph that are matched by target scope.
    If strict is True, raises nncf.ValidationError if no rule is matched.

    :param target_scope: Target scope specifying the matching rules.
    :param nncf_graph: The graph.
    :param strict: Whether target_scope must match at least one node or not.
    :return: NNCF node names from the given graph matched by target scope.
    """
    matched_target_scope, matches = get_matched_ignored_scope_info(target_scope, [nncf_graph])
    if strict:
        _check_target_scope_strictly_matched(target_scope, matched_target_scope)
    return set().union(*matches.values())


def _check_target_scope_strictly_matched(target_scope: TargetScope, matched_target_scope: TargetScope):
    """
    Passes when target_scope and matched_target_scope are equal, otherwise raises ValidationError.

    :param target_scope: The given target scope.
    :param matched_target_scope: The actual target scope matched in a graph.
    """
    unmatched_scope = get_difference_ignored_scope(target_scope, matched_target_scope)
    error_messages = []
    for match_type in ("names", "types", "patterns", "subgraphs"):
        unmatched_rules = getattr(unmatched_scope, match_type)
        if unmatched_rules:
            error_messages.append(f"The following {match_type} are not found in the graph: {unmatched_rules}.")
    if error_messages:
        raise nncf.ValidationError("\n".join(error_messages))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, cast

import nncf
from nncf.common.engine import Engine
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.transformations.command_creation import CommandCreator
from nncf.common.tensor_statistics import aggregator
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.data.dataset import Dataset

TModel = TypeVar("TModel")


class NNCFGraphFactory:
    @staticmethod
    def create(model: TModel) -> NNCFGraph:
        """
        Factory method to create backend-specific NNCFGraph instance based on the input model.

        :param model: backend-specific model instance
        :return: backend-specific NNCFGraph instance
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.ONNX:
            from onnx import ModelProto  # type: ignore

            from nncf.onnx.graph.nncf_graph_builder import GraphConverter as ONNXGraphConverter

            return ONNXGraphConverter.create_nncf_graph(cast(ModelProto, model))
        if model_backend == BackendType.OPENVINO:
            from openvino import Model  # type: ignore

            from nncf.openvino.graph.nncf_graph_builder import GraphConverter as OVGraphConverter

            return OVGraphConverter.create_nncf_graph(cast(Model, model))
        if model_backend == BackendType.TORCH_FX:
            from torch.fx import GraphModule

            from nncf.experimental.torch.fx.nncf_graph_builder import GraphConverter as FXGraphConverter

            return FXGraphConverter.create_nncf_graph(cast(GraphModule, model))
        if model_backend == BackendType.TORCH:
            from nncf.torch.nncf_network import NNCFNetwork

            return cast(NNCFNetwork, model).nncf.get_graph()
        raise nncf.UnsupportedBackendError(
            "Cannot create backend-specific graph because {} is not supported!".format(model_backend.value)
        )


class ModelTransformerFactory:
    @staticmethod
    def create(model: TModel, inplace: bool = False) -> ModelTransformer:
        """
        Factory method to create backend-specific ModelTransformer instance based on the input model.

        :param model: backend-specific model instance
        :param inplace: apply transformations inplace
        :return: backend-specific ModelTransformer instance
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.ONNX:
            from onnx import ModelProto

            from nncf.onnx.graph.model_transformer import ONNXModelTransformer

            return ONNXModelTransformer(cast(ModelProto, model))
        if model_backend == BackendType.OPENVINO:
            from openvino import Model

            from nncf.openvino.graph.model_transformer import OVModelTransformer

            return OVModelTransformer(cast(Model, model), inplace=inplace)
        if model_backend == BackendType.TORCH:
            from nncf.torch.model_transformer import PTModelTransformer
            from nncf.torch.nncf_network import NNCFNetwork

            return PTModelTransformer(cast(NNCFNetwork, model))
        if model_backend == BackendType.TORCH_FX:
            from torch.fx import GraphModule

            from nncf.experimental.torch.fx.model_transformer import FXModelTransformer

            return FXModelTransformer(cast(GraphModule, model))
        raise nncf.UnsupportedBackendError(
            "Cannot create backend-specific model transformer because {} is not supported!".format(model_backend.value)
        )


class EngineFactory:
    @staticmethod
    def create(model: TModel) -> Engine:
        """
        Factory method to create backend-specific Engine instance based on the input model.

        :param model: backend-specific model instance.
        :return: backend-specific Engine instance.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.ONNX:
            from onnx import ModelProto

            from nncf.onnx.engine import ONNXEngine

            return ONNXEngine(cast(ModelProto, model))
        if model_backend == BackendType.OPENVINO:
            from openvino import Model

            from nncf.openvino.engine import OVNativeEngine

            return OVNativeEngine(cast(Model, model))
        if model_backend in (BackendType.TORCH, BackendType.TORCH_FX):
            from torch.nn import Module

            from nncf.torch.engine import PTEngine

            return PTEngine(cast(Module, model))
        raise nncf.UnsupportedBackendError(
            "Cannot create backend-specific engine because {} is not supported!".format(model_backend.value)
        )


class CommandCreatorFactory:
    @staticmethod
    def create(model: TModel) -> CommandCreator:
        """
        Factory method to create backend-specific `CommandCreator` instance based on the input model.

        :param model: backend-specific model instance
        :return: backend-specific CommandCreator instance
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.openvino.graph.transformations.command_creation import OVCommandCreator

            return OVCommandCreator()

        if model_backend == BackendType.ONNX:
            from nncf.onnx.graph.transformations.command_creation import ONNXCommandCreator

            return ONNXCommandCreator()

        raise nncf.UnsupportedBackendError(
            "Cannot create backend-specific command creator because {} is not supported!".format(model_backend.value)
        )


class StatisticsAggregatorFactory:
    @staticmethod
    def create(model: TModel, dataset: Dataset) -> aggregator.StatisticsAggregator:
        """
        Factory method to create backend-specific `StatisticsAggregator` instance based on the input model.

        :param model: backend-specific model instance
        :return: backend-specific `StatisticsAggregator` instance
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.ONNX:
            from nncf.onnx.statistics.aggregator import ONNXStatisticsAggregator

            return ONNXStatisticsAggregator(dataset)
        if model_backend == BackendType.OPENVINO:
            from nncf.openvino.statistics.aggregator import OVStatisticsAggregator

            return OVStatisticsAggregator(dataset)
        if model_backend == BackendType.TORCH:
            from nncf.torch.statistics.aggregator import PTStatisticsAggregator

            return PTStatisticsAggregator(dataset)
        if model_backend == BackendType.TORCH_FX:
            from nncf.experimental.torch.fx.statistics.aggregator import FXStatisticsAggregator

            return FXStatisticsAggregator(dataset)
        raise nncf.UnsupportedBackendError(
            "Cannot create backend-specific statistics aggregator because {} is not supported!".format(
                model_backend.value
            )
        )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Any


class Engine(ABC):
    """
    The basic class aims to provide the interface to infer the model.
    """

    @abstractmethod
    def infer(self, input_data: Any) -> Any:
        """
        Runs model on the provided input data.
        Returns the raw model outputs.

        :param input_data: inputs for the model
        :return: raw model outputs
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import numpy as np
import torch

from nncf.common.factory import TModel
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.tensor_statistics.aggregator import StatisticPointsContainer
from nncf.common.tensor_statistics.aggregator import StatisticsAggregator
from nncf.common.utils.backend import BackendType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import TensorStatistic
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.transformations import leaf_module_insertion_transformation_builder
from nncf.tensor import Tensor
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.return_types import maybe_get_values_from_torch_return_type


class TensorCollectorModule(torch.nn.Module):
    """
    torch.nn.Module which calls given collector in forward
    """

    def __init__(self, collector: TensorCollector):
        super().__init__()
        self._collector = collector

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Register inputs hook function.

        :parameter x: tensor to register in hook.
        :return: tensor to register in hook.
        """
        x_unwrapped = maybe_get_values_from_torch_return_type(x)
        self._collector.register_input_for_all_reducers(Tensor(x_unwrapped))
        return x


class FXStatisticsAggregator(StatisticsAggregator):
    BACKEND: BackendType = BackendType.TORCH_FX
    HOOKS_GROUP_NAME = "statistics_hooks"

    def collect_statistics(self, model: NNCFNetwork, graph: NNCFGraph) -> None:
        with torch.no_grad():
            super().collect_statistics(model, graph)
        # All statistics are collected as a dead code,
        # so eliminate dead core removed statistcs collector
        # from the target model. No additional code required
        # for that, horay!
        model.graph.eliminate_dead_code()
        model.recompile()

    def _register_statistics(self, outputs: Dict[str, Tensor], statistic_points: StatisticPointsContainer) -> None:
        return

    @staticmethod
    def _get_statistic_collector_name(tp: PTTargetPoint, module_to_insert: torch.nn.Module) -> str:
        """
        Compouses unique statistic collector name according to given target point and module.

        :param tp: Given target point.
        :param module_to_insert: Given statistic collection module.
        :return: Unique statistic collector name according to given target point and module.
        """
        return "_".join(
            [
                tp.target_node_name,
                str(tp.input_port_id),
                str(tp.target_type.value),
                str(id(module_to_insert)),
            ]
        )

    def _get_transformation_layout_extra_outputs(
        self, statistic_points: StatisticPointsContainer
    ) -> TransformationLayout:
        transformation_layout = TransformationLayout()
        transformation_commands = []

        for _statistic_points in statistic_points.values():
            for _statistic_point in _statistic_points:
                for collectors in _statistic_point.algorithm_to_tensor_collectors.values():
                    for collector in collectors:
                        tp = _statistic_point.target_point
                        module_to_insert = TensorCollectorModule(collector)
                        target_module_name = self._get_statistic_collector_name(tp, module_to_insert)
                        transformation = leaf_module_insertion_transformation_builder(
                            module_to_insert, [tp], target_module_name
                        )
                        transformation_commands.append(
                            FXApplyTransformationCommand(
                                transformation, TransformationPriority.FP32_TENSOR_STATISTICS_OBSERVATION
                            )
                        )

        for transformation_command in transformation_commands:
            transformation_layout.register(transformation_command)

        return transformation_layout

    @staticmethod
    def _get_merged_statistic_points(
        statistic_points: StatisticPointsContainer, model: TModel, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        # TODO(dlyakhov): mirgate to experimental statistic collector and use common merging algorithm
        return statistic_points

    @staticmethod
    def _process_outputs(outputs: Dict[str, np.ndarray]) -> Dict[str, Tensor]:
        return outputs

    def _get_statistics_key(self, statistics: TensorStatistic, target_point: PTTargetPoint) -> str:
        """
        Returns key of statistics.

        :param statistics: Statistics value.
        :param target_point: Statistics target point.
        :return: Statistics key.
        """
        target_point_id = f"{target_point.target_node_name}_{target_point.type}_{target_point.input_port_id}"
        return f"{statistics.__class__.__name__}_{target_point_id}"
